"""
Contains the core building blocks of the framework.
"""

import math
from copy import deepcopy

import cython as cy
import numpy as np
import pandas as pd

PAR = 100.0
TOL = 1e-16


@cy.locals(x=cy.double)
def is_zero(x):
    """
    Test for zero that is robust against floating point precision errors
    """
    return abs(x) < TOL


class Node(object):
    """
    The Node is the main building block in bt's tree structure design.
    Both StrategyBase and SecurityBase inherit Node. It contains the
    core functionality of a tree node.

    Args:
        * name (str): The Node name
        * parent (Node): The parent Node
        * children (dict, list): A collection of children. If dict,
          the format is {name: child}, if list then list of children.
          Children can be any type of Node or str.
          String values correspond to children which will be lazily created
          with that name when needed.

    Attributes:
        * name (str): Node name
        * parent (Node): Node parent
        * root (Node): Root node of the tree (topmost node)
        * children (dict): Node's children
        * now (datetime): Used when backtesting to store current date
        * stale (bool): Flag used to determine if Node is stale and need
          updating
        * prices (TimeSeries): Prices of the Node. Prices for a security will
          be the security's price, for a strategy it will be an index that
          reflects the value of the strategy over time.
        * price (float): last price
        * value (float): last value
        * notional_value (float): last notional value. Notional value is used
          when fixed_income=True. It is always positive for strategies, but
          is signed for securities (and typically set to either market value,
          position, or zero).
        * weight (float): weight in parent
        * full_name (str): Name including parents' names
        * members (list): Current Node + node's children
        * fixed_income (bool): Whether the node corresponds to a fixed income
          component, which would use notional-weighting instead of market
          value weighing. See also :class:`FixedIncomeStrategy <bt.core.FixedIncomeStrategy>`
          for more details.
    """

    _capital = cy.declare(cy.double)
    _price = cy.declare(cy.double)
    _value = cy.declare(cy.double)
    _notl_value = cy.declare(cy.double)
    _weight = cy.declare(cy.double)
    _issec = cy.declare(cy.bint)
    _has_strat_children = cy.declare(cy.bint)
    _fixed_income = cy.declare(cy.bint)
    _bidoffer_set = cy.declare(cy.bint)
    _bidoffer_paid = cy.declare(cy.double)

    def __init__(self, name, parent=None, children=None):
        self.name = name

        # children helpers
        self.children = {}
        self._lazy_children = {}
        self._universe_tickers = []
        self._childrenv = []  # Shortcut to self.children.values()
        self._original_children_are_present = (children is not None) and (len(children) >= 1)

        # strategy children helpers
        self._has_strat_children = False
        self._strat_children = []

        if parent is None:
            self.parent = self
            self.root = self
            # by default all positions are integer
            self.integer_positions = True
        else:
            self.parent = parent
            parent._add_children([self], dc=False)

        self._add_children(children, dc=True)

        # set default value for now
        self.now = 0
        # make sure root has stale flag
        # used to avoid unnecessary update
        # sometimes we change values in the tree and we know that we will need
        # to update if another node tries to access a given value (say weight).
        # This avoid calling the update until it is actually needed.
        self.root.stale = False

        # helper vars
        self._price = 0
        self._value = 0
        self._notl_value = 0
        self._weight = 0
        self._capital = 0

        # is security flag - used to avoid updating 0 pos securities
        self._issec = False

        # fixed income flag - used to turn on notional weighing
        self._fixed_income = False
        # flag for whether to do bid/offer accounting
        self._bidoffer_set = False
        self._bidoffer_paid = 0

    def __getitem__(self, key):
        return self.children[key]

    def _add_children(self, children, dc):
        """
        Add the collection of children to the current node, where
        children is either an iterable of children objects/strings, or
        a dictionary

        Args:
            dc (bool): Whether or not to deepcopy nodes before adding them.
        """
        # if at least 1 children is specified
        if children is not None:
            if isinstance(children, dict):
                # Preserve the names from the dictionary by renaming the nodes
                tmp = []
                for name, c in children.items():
                    if isinstance(c, str):
                        tmp.append(name)
                    else:
                        if dc:
                            c = deepcopy(c)
                        c.name = name
                        tmp.append(c)
                children = tmp

            for c in children:
                if dc:  # deepcopy object for possible later reuse
                    c = deepcopy(c)

                if isinstance(c, str):
                    if c in self._universe_tickers:
                        raise ValueError("Child %s already exists" % c)

                    # Create default security with lazy_add
                    c = Security(c, lazy_add=True)

                if getattr(c, "lazy_add", False):
                    self._lazy_children[c.name] = c
                else:
                    if c.name in self.children:
                        raise ValueError("Child %s already exists" % c)

                    c.parent = self
                    c._set_root(self.root)
                    c.use_integer_positions(self.integer_positions)

                    self.children[c.name] = c
                    self._childrenv.append(c)

                # if strategy, turn on flag and add name to list
                # strategy children have special treatment
                if isinstance(c, StrategyBase):
                    self._has_strat_children = True
                    self._strat_children.append(c.name)
                # if not strategy, then we will want to add this to
                # universe_tickers to filter on setup
                elif c.name not in self._universe_tickers:
                    self._universe_tickers.append(c.name)

    def _set_root(self, root):
        self.root = root
        for c in self._childrenv:
            c._set_root(root)

    def use_integer_positions(self, integer_positions):
        """
        Set indicator to use (or not) integer positions for a given strategy or
        security.

        By default all positions in number of stocks should be integer.
        However this may lead to unexpected results when working with adjusted
        prices of stocks. Because of series of reverse splits of stocks, the
        adjusted prices back in time might be high. Thus rounding of desired
        amount of stocks to buy may lead to having 0, and thus ignoring this
        stock from backtesting.
        """
        self.integer_positions = integer_positions
        for c in self._childrenv:
            c.use_integer_positions(integer_positions)

    @property
    def fixed_income(self):
        """
        Whether the node is a fixed income node (using notional weighting).
        """
        return self._fixed_income

    @property
    def prices(self):
        """
        A TimeSeries of the Node's price.
        """
        # can optimize depending on type -
        # securities don't need to check stale to
        # return latest prices, whereas strategies do...
        raise NotImplementedError()

    @property
    def price(self):
        """
        Current price of the Node
        """
        # can optimize depending on type -
        # securities don't need to check stale to
        # return latest prices, whereas strategies do...
        raise NotImplementedError()

    @property
    def value(self):
        """
        Current value of the Node
        """
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._value

    @property
    def notional_value(self):
        """
        Current notional value of the Node
        """
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._notl_value

    @property
    def weight(self):
        """
        Current weight of the Node (with respect to the parent).
        """
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._weight

    def setup(self, universe, **kwargs):
        """
        Setup method used to initialize a Node with a universe, and potentially other information.
        """
        raise NotImplementedError()

    def update(self, date, data=None, inow=None):
        """
        Update Node with latest date, and optionally some data.
        """
        raise NotImplementedError()

    def adjust(self, amount, update=True, flow=True):
        """
        Adjust Node value by amount.
        """
        raise NotImplementedError()

    def allocate(self, amount, update=True):
        """
        Allocate capital to Node.
        """
        raise NotImplementedError()

    @property
    def members(self):
        """
        Node members. Members include current node as well as Node's
        children.
        """
        res = [self]
        for c in list(self.children.values()):
            res.extend(c.members)
        return res

    @property
    def full_name(self):
        if self.parent == self:
            return self.name
        else:
            return "%s>%s" % (self.parent.full_name, self.name)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.full_name)

    def to_dot(self, root=True):
        """
        Represent the node structure in DOT format.
        """
        name = lambda x: x.name or repr(self)  # noqa: E731
        edges = "\n".join('\t"%s" -> "%s"' % (name(self), name(c)) for c in self.children.values())
        below = "\n".join(c.to_dot(False) for c in self.children.values())
        body = "\n".join([edges, below]).rstrip()
        if root:
            return "\n".join(["digraph {", body, "}"])
        return body


class StrategyBase(Node):
    """
    Strategy Node. Used to define strategy logic within a tree.
    A Strategy's role is to allocate capital to it's children
    based on a function.

    Args:
        * name (str): Strategy name
        * children (dict, list): A collection of children. If dict,
          the format is {name: child}, if list then list of children.
          Children can be any type of Node or str.
          String values correspond to children which will be lazily created
          with that name when needed.
        * parent (Node): The parent Node

    Attributes:
        * name (str): Strategy name
        * parent (Strategy): Strategy parent
        * root (Strategy): Root node of the tree (topmost node)
        * children (dict): Strategy's children
        * now (datetime): Used when backtesting to store current date
        * stale (bool): Flag used to determine if Strategy is stale and need
          updating
        * prices (TimeSeries): Prices of the Strategy - basically an index that
          reflects the value of the strategy over time.
        * outlays (DataFrame): Outlays for each SecurityBase child
        * price (float): last price
        * value (float): last value
        * notional_value (float): last notional value
        * weight (float): weight in parent
        * full_name (str): Name including parents' names
        * members (list): Current Strategy + strategy's children
        * securities (list): List of strategy children that are of type
          SecurityBase
        * commission_fn (fn(quantity, price)): A function used to determine the
          commission (transaction fee) amount. Could be used to model
          slippage (implementation shortfall). Note that often fees are
          symmetric for buy and sell and absolute value of quantity should
          be used for calculation.
        * capital (float): Capital amount in Strategy - cash
        * universe (DataFrame): Data universe available at the current time.
          Universe contains the data passed in when creating a Backtest. Use
          this data to determine strategy logic.

    """

    _net_flows = cy.declare(cy.double)
    _last_value = cy.declare(cy.double)
    _last_notl_value = cy.declare(cy.double)
    _last_price = cy.declare(cy.double)
    _last_fee = cy.declare(cy.double)
    _paper_trade = cy.declare(cy.bint)
    bankrupt = cy.declare(cy.bint)
    _last_chk = cy.declare(cy.bint)

    def __init__(self, name, children=None, parent=None):
        Node.__init__(self, name, children=children, parent=parent)
        self._weight = 1
        self._value = 0
        self._notl_value = 0
        self._price = PAR

        # helper vars
        self._net_flows = 0
        self._last_value = 0
        self._last_notl_value = 0
        self._last_price = PAR
        self._last_fee = 0

        self._last_chk = 0

        # default commission function
        self.commission_fn = self._dflt_comm_fn

        self._paper_trade = False
        self._positions = None
        self.bankrupt = False

    @property
    def price(self):
        """
        Current price.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._price

    @property
    def prices(self):
        """
        TimeSeries of prices.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._prices.loc[: self.now]

    @property
    def values(self):
        """
        TimeSeries of values.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._values.loc[: self.now]

    @property
    def notional_values(self):
        """
        TimeSeries of notional values.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._notl_values.loc[: self.now]

    @property
    def capital(self):
        """
        Current capital - amount of unallocated capital left in strategy.
        """
        # no stale check needed
        return self._capital

    @property
    def cash(self):
        """
        TimeSeries of unallocated capital.
        """
        # no stale check needed
        return self._cash

    @property
    def fees(self):
        """
        TimeSeries of fees.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._fees.loc[: self.now]

    @property
    def flows(self):
        """
        TimeSeries of flows.
        """
        if self.root.stale:
            self.root.update(self.now, None)
        return self._all_flows.loc[: self.now]

    @property
    def bidoffer_paid(self):
        """
        Bid/offer spread paid on transactions in the current step
        """
        if self._bidoffer_set:
            if self.root.stale:
                self.root.update(self.now, None)
            return self._bidoffer_paid
        else:
            raise Exception("Bid/offer accounting not turned on: " '"bidoffer" argument not provided during setup')

    @property
    def bidoffers_paid(self):
        """
        TimeSeries of bid/offer spread paid on transactions in each step
        """
        if self._bidoffer_set:
            if self.root.stale:
                self.root.update(self.now, None)
            return self._bidoffers_paid.loc[: self.now]
        else:
            raise Exception("Bid/offer accounting not turned on: " '"bidoffer" argument not provided during setup')

    @property
    def universe(self):
        """
        Data universe available at the current time.
        Universe contains the data passed in when creating a Backtest.
        Use this data to determine strategy logic.
        """
        # avoid windowing every time
        # if calling and on same date return
        # cached value
        if self.now == self._last_chk:
            return self._funiverse
        else:
            self._last_chk = self.now
            self._funiverse = self._universe.loc[: self.now]
            return self._funiverse

    @property
    def securities(self):
        """
        Returns a list of children that are of type SecurityBase
        """
        return [x for x in self.members if isinstance(x, SecurityBase)]

    @property
    def outlays(self):
        """
        Returns a DataFrame of outlays for each child SecurityBase
        """
        if self.root.stale:
            self.root.update(self.root.now, None)
        outlays = pd.DataFrame()
        for x in self.securities:
            if x.name in outlays.columns:
                outlays[x.name] += x.outlays
            else:
                outlays[x.name] = x.outlays
        return outlays

    @property
    def positions(self):
        """
        TimeSeries of positions.
        """
        # if accessing and stale - update first
        if self.root.stale:
            self.root.update(self.root.now, None)

        vals = pd.DataFrame()
        for x in self.members:
            if isinstance(x, SecurityBase):
                if x.name in vals.columns:
                    vals[x.name] += x.positions
                else:
                    vals[x.name] = x.positions
        self._positions = vals.fillna(0.0)
        return vals

    def setup(self, universe, **kwargs):
        """
        Setup strategy with universe. This will speed up future calculations
        and updates.
        """
        # save full universe in case we need it
        self._original_data = universe
        self._setup_kwargs = kwargs

        # Guard against fixed income children of regular
        # strategies as the "price" is just a reference
        # value and should not be used for capital allocation
        if self.fixed_income and not self.parent.fixed_income:
            raise ValueError("Cannot have fixed income " "strategy child (%s) of non-" "fixed income strategy (%s)" % (self.name, self.parent.name))

        # determine if needs paper trading
        # and setup if so
        if self is not self.parent:
            self._paper_trade = True
            self._paper_amount = 1000000

            paper = deepcopy(self)
            paper.parent = paper
            paper.root = paper
            paper._paper_trade = False
            paper.setup(self._original_data, **kwargs)
            paper.adjust(self._paper_amount)
            self._paper = paper

        # setup universe
        funiverse = universe.copy()

        # filter only if the node has any children specified as input,
        # otherwise we use the full universe. If all children are strategies,
        # funiverse will be empty, to signal that no other ticker should be
        # used in addition to the strategies
        if self._original_children_are_present:
            # if we have universe_tickers defined, limit universe to
            # those tickers
            valid_filter = list(set(universe.columns).intersection(self._universe_tickers))

            funiverse = universe[valid_filter].copy()

            # if we have strat children, we will need to create their columns
            # in the new universe
            if self._has_strat_children:
                for c in self._strat_children:
                    funiverse[c] = np.nan

            # must create to avoid pandas warning
            funiverse = pd.DataFrame(funiverse)

        self._universe = funiverse
        # holds filtered universe
        self._funiverse = funiverse
        self._last_chk = None

        # We're not bankrupt yet
        self.bankrupt = False

        # setup internal data
        self.data = pd.DataFrame(
            index=funiverse.index,
            columns=["price", "value", "notional_value", "cash", "fees", "flows"],
            data=0.0,
        )

        self._prices = self.data["price"]
        self._values = self.data["value"]
        self._notl_values = self.data["notional_value"]
        self._cash = self.data["cash"]
        self._fees = self.data["fees"]
        self._all_flows = self.data["flows"]

        if "bidoffer" in kwargs:
            self._bidoffer_set = True
            self.data["bidoffer_paid"] = 0.0
            self._bidoffers_paid = self.data["bidoffer_paid"]

        # setup children as well - use original universe here - don't want to
        # pollute with potential strategy children in funiverse
        if self.children is not None:
            [c.setup(universe, **kwargs) for c in self._childrenv]

    def setup_from_parent(self, **kwargs):
        """
        Setup a strategy from the parent. Used when dynamically creating
        child strategies.

        Args:
            * kwargs: additional arguments that will be passed to setup
                (potentially overriding those from the parent)
        """
        all_kwargs = self.parent._setup_kwargs.copy()
        all_kwargs.update(kwargs)
        self.setup(self.parent._original_data, **all_kwargs)
        if self.name not in self.parent._universe:
            self.parent._universe[self.name] = np.nan

    def get_data(self, key):
        """
        Returns additional data that was passed to the setup function via kwargs,
        for use in the algos. This allows algos to reference data sources "by name",
        where the binding of the data to the name happens at Backtest creation
        time rather than at Strategy definition time, allowing the same strategies
        to be run against different data sets more easily.
        """
        return self._setup_kwargs[key]

    @cy.locals(
        newpt=cy.bint,
        val=cy.double,
        ret=cy.double,
        coupons=cy.double,
        notl_val=cy.double,
        bidoffer_paid=cy.double,
    )
    def update(self, date, data=None, inow=None):
        """
        Update strategy. Updates prices, values, weight, etc.
        """
        # resolve stale state
        self.root.stale = False

        # update helpers on date change
        # also set newpt flag
        newpt = False
        if self.now == 0:
            newpt = True
        elif date != self.now:
            self._net_flows = 0
            self._last_price = self._price
            self._last_value = self._value
            self._last_notl_value = self._notl_value
            self._last_fee = 0.0
            newpt = True

        # update now
        self.now = date
        if inow is None:
            if self.now == 0:
                inow = 0
            else:
                inow = self.data.index.get_loc(date)

        # update children if any and calculate value
        val = self._capital  # default if no children
        notl_val = 0.0  # Capital doesn't count towards notional value

        bidoffer_paid = 0.0
        coupons = 0
        if self.children:
            for c in self._childrenv:
                # Sweep up cash from the security nodes (from coupon payments, etc)
                if c._issec and newpt:
                    coupons += c._capital
                    c._capital = 0

                # avoid useless update call
                if c._issec and not c._needupdate:
                    continue

                c.update(date, data, inow)
                val += c.value
                # Strategies always have positive notional value
                notl_val += abs(c.notional_value)

                if self._bidoffer_set:
                    bidoffer_paid += c.bidoffer_paid

        self._capital += coupons
        val += coupons

        if self.root == self:
            if (val < 0) and not self.bankrupt and not self.fixed_income and not is_zero(val):
                # Declare a bankruptcy
                self.bankrupt = True
                self.flatten()

        # update data if this value is different or
        # if now has changed - avoid all this if not since it
        # won't change
        if newpt or not is_zero(self._value - val) or not is_zero(self._notl_value - notl_val):
            self._value = val
            self._values.values[inow] = val

            self._notl_value = notl_val
            self._notl_values.values[inow] = notl_val

            if self._bidoffer_set:
                self._bidoffer_paid = bidoffer_paid
                self._bidoffers_paid.values[inow] = bidoffer_paid

            if self.fixed_income:
                # For notional weights, we compute additive return
                pnl = self._value - (self._last_value + self._net_flows)
                if not is_zero(self._last_notl_value):
                    ret = pnl / self._last_notl_value * PAR
                elif not is_zero(self._notl_value):
                    # This case happens when paying bid/offer or fees when building an initial position
                    ret = pnl / self._notl_value * PAR
                else:
                    if is_zero(pnl):
                        ret = 0
                    else:
                        raise ZeroDivisionError(
                            "Could not update %s on %s. Last notional value "
                            "was %s and pnl was %s. Therefore, "
                            "we are dividing by zero to obtain the pnl "
                            "per unit notional for the period." % (self.name, self.now, self._last_notl_value, pnl)
                        )

                self._price = self._last_price + ret
                self._prices.values[inow] = self._price

            else:
                bottom = self._last_value + self._net_flows
                if not is_zero(bottom):
                    ret = self._value / (self._last_value + self._net_flows) - 1
                else:
                    if is_zero(self._value):
                        ret = 0
                    else:
                        raise ZeroDivisionError(
                            "Could not update %s on %s. Last value "
                            "was %s and net flows were %s. Current"
                            "value is %s. Therefore, "
                            "we are dividing by zero to obtain the return "
                            "for the period."
                            % (
                                self.name,
                                self.now,
                                self._last_value,
                                self._net_flows,
                                self._value,
                            )
                        )

                self._price = self._last_price * (1 + ret)
                self._prices.values[inow] = self._price

        # update children weights
        if self.children:
            for c in self._childrenv:
                # avoid useless update call
                if c._issec and not c._needupdate:
                    continue

                if self.fixed_income:
                    if not is_zero(notl_val):
                        c._weight = c.notional_value / notl_val
                    else:
                        c._weight = 0.0
                else:
                    if not is_zero(val):
                        c._weight = c.value / val
                    else:
                        c._weight = 0.0

        # if we have strategy children, we will need to update them in universe
        if self._has_strat_children:
            for c in self._strat_children:
                # TODO: optimize ".loc" here as well
                self._universe.loc[date, c] = self.children[c].price

        # Cash should track the unallocated capital at the end of the day, so
        # we should update it every time we call "update".
        # Same for fees and flows
        self._cash.values[inow] = self._capital
        self._fees.values[inow] = self._last_fee
        self._all_flows.values[inow] = self._net_flows

        # update paper trade if necessary
        if self._paper_trade:
            if newpt:
                self._paper.update(date)
                self._paper.run()
                self._paper.update(date)
            # update price
            self._price = self._paper.price
            self._prices.values[inow] = self._price

    @cy.locals(amount=cy.double, update=cy.bint, flow=cy.bint, fees=cy.double)
    def adjust(self, amount, update=True, flow=True, fee=0.0):
        """
        Adjust capital - used to inject capital to a Strategy. This injection
        of capital will have no effect on the children.

        Args:
            * amount (float): Amount to adjust by.
            * update (bool): Force update?
            * flow (bool): Is this adjustment a flow? A flow will not have an
              impact on the performance (price index). Example of flows are
              simply capital injections (say a monthly contribution to a
              portfolio). This should not be reflected in the returns. A
              non-flow (flow=False) does impact performance. A good example
              of this is a commission, or a dividend.

        """
        # adjust capital
        self._capital += amount
        self._last_fee += fee

        # if flow - increment net_flows - this will not affect
        # performance. Commissions and other fees are not flows since
        # they have a performance impact
        if flow:
            self._net_flows += amount

        if update:
            # indicates that data is now stale and must
            # be updated before access
            self.root.stale = True

    @cy.locals(amount=cy.double, update=cy.bint)
    def allocate(self, amount, child=None, update=True):
        """
        Allocate capital to Strategy. By default, capital is allocated
        recursively down the children, proportionally to the children's
        weights.  If a child is specified, capital will be allocated
        to that specific child.

        Allocation also have a side-effect. They will deduct the same amount
        from the parent's "account" to offset the allocation. If there is
        remaining capital after allocation, it will remain in Strategy.

        Args:
            * amount (float): Amount to allocate.
            * child (str): If specified, allocation will be directed to child
              only. Specified by name.
            * update (bool): Force update.

        """
        # allocate to child
        if child is not None:
            self._create_child_if_needed(child)

            # allocate to child
            self.children[child].allocate(amount)
        # allocate to self
        else:
            # adjust parent's capital
            # no need to update now - avoids repetition
            if self.parent == self:
                self.parent.adjust(-amount, update=False, flow=True)
            else:
                # do NOT set as flow - parent will be another strategy
                # and therefore should not incur flow
                self.parent.adjust(-amount, update=False, flow=False)

            # adjust self's capital
            self.adjust(amount, update=False, flow=True)

            # push allocation down to children if any
            # use _weight to avoid triggering an update
            if self.children is not None:
                [c.allocate(amount * c._weight, update=False) for c in self._childrenv]

            # mark as stale if update requested
            if update:
                self.root.stale = True

    @cy.locals(q=cy.double, update=cy.bint)
    def transact(self, q, child=None, update=True):
        """
        Transact a notional amount q in the Strategy. By default, it is allocated
        recursively down the children, proportionally to the children's
        weights. Recursive allocation only works for fixed income strategies.
        If a child is specified, notional will be allocated
        to that specific child.

        Args:
            * q (float): Notional quantity to allocate.
            * child (str): If specified, allocation will be directed to child
              only. Specified by name.
            * update (bool): Force update.

        """
        # allocate to child
        if child is not None:
            self._create_child_if_needed(child)

            # allocate to child
            self.children[child].transact(q)
        # allocate to self
        else:
            # push allocation down to children if any
            # use _weight to avoid triggering an update
            if self.children is not None:
                [c.transact(q * c._weight, update=False) for c in self._childrenv]

            # mark as stale if update requested
            if update:
                self.root.stale = True

    @cy.locals(delta=cy.double, weight=cy.double, base=cy.double, update=cy.bint)
    def rebalance(self, weight, child, base=np.nan, update=True):
        """
        Rebalance a child to a given weight.

        This is a helper method to simplify code logic. This method is used
        when we want to see the weight of a particular child to a set amount.
        It is similar to allocate, but it calculates the appropriate allocation
        based on the current weight. For fixed income strategies, it uses
        transact to rebalance based on notional value instead of capital.

        Args:
            * weight (float): The target weight. Usually between -1.0 and 1.0.
            * child (str): child to allocate to - specified by name.
            * base (float): If specified, this is the base amount all weight
              delta calculations will be based off of. This is useful when we
              determine a set of weights and want to rebalance each child
              given these new weights. However, as we iterate through each
              child and call this method, the base (which is by default the
              current value) will change. Therefore, we can set this base to
              the original value before the iteration to ensure the proper
              allocations are made.
            * update (bool): Force update?

        """
        # if weight is 0 - we want to close child
        if is_zero(weight):
            if child in self.children:
                return self.close(child, update=update)
            else:
                return

        # if no base specified use self's value
        if np.isnan(base):
            if self.fixed_income:
                base = self.notional_value
            else:
                base = self.value

        # else make sure we have child
        self._create_child_if_needed(child)

        # allocate to child
        # figure out weight delta
        c = self.children[child]

        if self.fixed_income:
            # In fixed income strategies, the provided "base" value can be used
            # to upscale/downscale the notional_value of the strategy, whereas
            # in normal strategies the total capital is fixed. Thus, when
            # rebalancing, we must take care to account for differences between
            # previous notional value and passed base value. Note that for
            # updating many weights in sequence, one must pass update=False so
            # that the existing weights and notional_value are not recalculated
            # before finishing.
            if c.fixed_income:
                delta = weight * base - c.weight * self.notional_value
                c.transact(delta, update=update)
            else:
                delta = weight * base - c.weight * self.notional_value
                c.allocate(delta, update=update)
        else:
            delta = weight - c.weight
            c.allocate(delta * base, update=update)

    @cy.locals(update=cy.bint)
    def close(self, child, update=True):
        """
        Close a child position - alias for rebalance(0, child). This will also
        flatten (close out all) the child's children.

        Args:
            * child (str): Child, specified by name.
        """
        c = self.children[child]
        # flatten if children not None
        if c.children is not None and len(c.children) != 0:
            c.flatten()

        if self.fixed_income:
            if c.position != 0.0:
                c.transact(-c.position, update=update)
        else:
            if c.value != 0.0 and not np.isnan(c.value):
                c.allocate(-c.value, update=update)

    def flatten(self):
        """
        Close all child positions.
        """
        # go right to base alloc
        if self.fixed_income:
            [c.transact(-c.position, update=False) for c in self._childrenv if c.position != 0]
        else:
            [c.allocate(-c.value, update=False) for c in self._childrenv if c.value != 0]

        self.root.stale = True

    def run(self):
        """
        This is the main logic method. Override this method to provide some
        algorithm to execute on each date change. This method is called by
        backtester.
        """
        pass

    def set_commissions(self, fn):
        """
        Set commission (transaction fee) function.

        Args:
            fn (fn(quantity, price)): Function used to determine commission
            amount.

        """
        self.commission_fn = fn

        for c in self._childrenv:
            if isinstance(c, StrategyBase):
                c.set_commissions(fn)

    def get_transactions(self):
        """
        Helper function that returns the transactions in the following format:

            Date, Security | quantity, price

        The result is a MultiIndex DataFrame.
        """
        # get prices for each security in the strategy & create unstacked
        # series
        prc = pd.DataFrame({x.name: x.prices for x in self.securities}).unstack()

        # get security positions
        positions = pd.DataFrame()
        for x in self.securities:
            if x.name in positions.columns:
                positions[x.name] += x.positions
            else:
                positions[x.name] = x.positions
        # trades are diff
        trades = positions.diff()
        # must adjust first row
        trades.iloc[0] = positions.iloc[0]
        # now convert to unstacked series, dropping nans along the way
        trades = trades[trades != 0].unstack().dropna()

        # Adjust prices for bid/offer paid if needed
        if self._bidoffer_set:
            bidoffer = pd.DataFrame({x.name: x.bidoffers_paid for x in self.securities}).unstack()
            prc += bidoffer / trades

        res = pd.DataFrame({"price": prc, "quantity": trades}).dropna(subset=["quantity"])

        # set names
        res.index.names = ["Security", "Date"]

        # swap levels so that we have (date, security) as index and sort
        res = res.swaplevel().sort_index()

        return res

    @cy.locals(q=cy.double, p=cy.double)
    def _dflt_comm_fn(self, q, p):
        return 0.0

    def _create_child_if_needed(self, child):
        if child not in self.children:
            # Look up name in lazy children, or create a default security
            c = self._lazy_children.pop(child, Security(child))
            c.lazy_add = False

            # add child to tree
            self._add_children([c], dc=False)
            c.setup(self._universe, **self._setup_kwargs)

            # update to bring up to speed
            c.update(self.now)


class SecurityBase(Node):
    """
    Security Node. Used to define a security within a tree.
    A Security's has no children. It simply models an asset that can be bought
    or sold.

    Args:
        * name (str): Security name
        * multiplier (float): security multiplier - typically used for
          derivatives or to trade in lots. The quantity of the Security will
          always be multiplied by this to determine the underlying amount.
        * lazy_add (bool): Flag to control whether instrument should be added
          to strategy children lazily, i.e. only when there is a transaction
          on the instrument. This improves performance of strategies which
          transact on a sparse set of children.

    Attributes:
        * name (str): Security name
        * parent (Security): Security parent
        * root (Security): Root node of the tree (topmost node)
        * now (datetime): Used when backtesting to store current date
        * stale (bool): Flag used to determine if Security is stale and need
          updating
        * prices (TimeSeries): Security prices.
        * price (float): last price
        * outlays (TimeSeries): Series of outlays. Positive outlays mean
          capital was allocated to security and security consumed that
          amount.  Negative outlays are the opposite. This can be useful for
          calculating turnover at the strategy level.
        * value (float): last value - basically position * price * multiplier
        * weight (float): weight in parent
        * full_name (str): Name including parents' names
        * members (list): Current Security + strategy's children
        * position (float): Current position (quantity).
        * bidoffer (float): Current bid/offer spread
        * bidoffers (TimeSeries): Series of bid/offer spreads
        * bidoffer_paid (TimeSeries): Series of bid/offer paid on transactions
    """

    _last_pos = cy.declare(cy.double)
    _position = cy.declare(cy.double)
    multiplier = cy.declare(cy.double)
    _prices_set = cy.declare(cy.bint)
    _needupdate = cy.declare(cy.bint)
    _outlay = cy.declare(cy.double)
    _bidoffer = cy.declare(cy.double)

    @cy.locals(multiplier=cy.double)
    def __init__(self, name, multiplier=1, lazy_add=False):
        Node.__init__(self, name, parent=None, children=None)
        self._value = 0
        self._price = 0
        self._weight = 0
        self._position = 0
        self.multiplier = multiplier
        self.lazy_add = lazy_add

        # opt
        self._last_pos = 0
        self._issec = True
        self._needupdate = True
        self._outlay = 0
        self._bidoffer = 0

    @property
    def price(self):
        """
        Current price.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        return self._price

    @property
    def prices(self):
        """
        TimeSeries of prices.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        return self._prices.loc[: self.now]

    @property
    def values(self):
        """
        TimeSeries of values.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._values.loc[: self.now]

    @property
    def notional_values(self):
        """
        TimeSeries of notional values.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._notl_values.loc[: self.now]

    @property
    def position(self):
        """
        Current position
        """
        # no stale check needed
        return self._position

    @property
    def positions(self):
        """
        TimeSeries of positions.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._positions.loc[: self.now]

    @property
    def outlays(self):
        """
        TimeSeries of outlays. Positive outlays (buys) mean this security
        received and consumed capital (capital was allocated to it). Negative
        outlays are the opposite (the security close/sold, and returned capital
        to parent).
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        if self.root.stale:
            self.root.update(self.root.now, None)
        return self._outlays.loc[: self.now]

    @property
    def bidoffer(self):
        """
        Current bid/offer spread.
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        return self._bidoffer

    @property
    def bidoffers(self):
        """
        TimeSeries of bid/offer spread
        """
        if self._bidoffer_set:
            # if accessing and stale - update first
            if self._needupdate or self.now != self.parent.now:
                self.update(self.root.now)
            return self._bidoffers.loc[: self.now]
        else:
            raise Exception("Bid/offer accounting not turned on: " '"bidoffer" argument not provided during setup')

    @property
    def bidoffer_paid(self):
        """
        TimeSeries of bid/offer spread paid on transactions in the current step
        """
        # if accessing and stale - update first
        if self._needupdate or self.now != self.parent.now:
            self.update(self.root.now)
        return self._bidoffer_paid

    @property
    def bidoffers_paid(self):
        """
        TimeSeries of bid/offer spread paid on transactions in the current step
        """
        if self._bidoffer_set:
            # if accessing and stale - update first
            if self._needupdate or self.now != self.parent.now:
                self.update(self.root.now)
            if self.root.stale:
                self.root.update(self.root.now, None)
            return self._bidoffers_paid.loc[: self.now]
        else:
            raise Exception("Bid/offer accounting not turned on: " '"bidoffer" argument not provided during setup')

    def setup(self, universe, **kwargs):
        """
        Setup Security with universe. Speeds up future runs.

        Args:
            * universe (DataFrame): DataFrame of prices with security's name as
              one of the columns.
            * bidoffer (DataFrame): Optional argument that represents the
              bid/offer spread on each security across time. If provided, the
              strategy will account for these costs when rebalancing.
            * kwargs (dict): Dictionary of additional information needed by
              the strategy. In particular, often takes the form of a DataFrame
              of security level information (i.e. signals, risk, etc).
        """
        # if we already have all the prices, we will store them to speed up
        # future updates
        try:
            prices = universe[self.name]
        except KeyError:
            prices = None

        # setup internal data
        if prices is not None:
            self._prices = prices
            self.data = pd.DataFrame(
                index=universe.index,
                columns=["value", "position", "notional_value"],
                data=0.0,
            )
            self._prices_set = True
        else:
            self.data = pd.DataFrame(
                index=universe.index,
                columns=["price", "value", "position", "notional_value"],
            )
            self._prices = self.data["price"]
            self._prices_set = False

        self._values = self.data["value"]
        self._notl_values = self.data["notional_value"]
        self._positions = self.data["position"]

        # add _outlay
        self.data["outlay"] = 0.0
        self._outlays = self.data["outlay"]

        # save bidoffer, if provided
        if "bidoffer" in kwargs:
            self._bidoffer_set = True
            self._bidoffers = kwargs["bidoffer"]
            try:
                bidoffers = self._bidoffers[self.name]
            except KeyError:
                bidoffers = None

            if bidoffers is not None:
                if bidoffers.index.equals(universe.index):
                    self._bidoffers = bidoffers
                else:
                    raise ValueError("Index of bidoffer must match universe data")
            else:
                self.data["bidoffer"] = 0.0
                self._bidoffers = self.data["bidoffer"]

            self.data["bidoffer_paid"] = 0.0
            self._bidoffers_paid = self.data["bidoffer_paid"]

    @cy.locals(prc=cy.double)
    def update(self, date, data=None, inow=None):
        """
        Update security with a given date and optionally, some data.
        This will update price, value, weight, etc.
        """
        # filter for internal calls when position has not changed - nothing to
        # do. Internal calls (stale root calls) have None data. Also want to
        # make sure date has not changed, because then we do indeed want to
        # update.
        if date == self.now and self._last_pos == self._position:
            return

        if inow is None:
            if date == 0:
                inow = 0
            else:
                inow = self.data.index.get_loc(date)

        # date change - update price
        if date != self.now:
            # update now
            self.now = date

            if self._prices_set:
                self._price = self._prices.values[inow]
            # traditional data update
            elif data is not None:
                prc = data[self.name]
                self._price = prc
                self._prices.values[inow] = prc

            # update bid/offer
            if self._bidoffer_set:
                self._bidoffer = self._bidoffers.values[inow]
                self._bidoffer_paid = 0.0

        self._positions.values[inow] = self._position
        self._last_pos = self._position

        if np.isnan(self._price):
            if is_zero(self._position):
                self._value = 0
            else:
                raise Exception("Position is open (non-zero: %s) and latest price is NaN " "for security %s on %s. Cannot update node value." % (self._position, self.name, date))
        else:
            self._value = self._position * self._price * self.multiplier

        self._notl_value = self._value

        self._values.values[inow] = self._value
        self._notl_values.values[inow] = self._notl_value

        if is_zero(self._weight) and is_zero(self._position):
            self._needupdate = False

        # save outlay to outlays
        if self._outlay != 0:
            self._outlays.values[inow] += self._outlay
            # reset outlay back to 0
            self._outlay = 0

        if self._bidoffer_set:
            self._bidoffers_paid.values[inow] = self._bidoffer_paid

    @cy.locals(amount=cy.double, update=cy.bint, q=cy.double, outlay=cy.double, i=cy.int)
    def allocate(self, amount, update=True):
        """
        This allocates capital to the Security. This is the method used to
        buy/sell the security.

        A given amount of shares will be determined on the current price, a
        commission will be calculated based on the parent's commission fn, and
        any remaining capital will be passed back up  to parent as an
        adjustment.

        Args:
            * amount (float): Amount of adjustment.
            * update (bool): Force update?

        """

        # will need to update if this has been idle for a while...
        # update if needupdate or if now is stale
        # fetch parent's now since our now is stale
        if self._needupdate or self.now != self.parent.now:
            self.update(self.parent.now)

        # ignore 0 alloc
        # Note that if the price of security has dropped to zero, then it
        # should never be selected by SelectAll, SelectN etc. I.e. we should
        # not open the position at zero price. At the same time, we are able
        # to close it at zero price, because at that point amount=0.
        # Note also that we don't erase the position in an asset which price
        # has dropped to zero (though the weight will indeed be = 0)
        if is_zero(amount):
            return

        if self.parent is self or self.parent is None:
            raise Exception("Cannot allocate capital to a parentless security")

        if is_zero(self._price) or np.isnan(self._price):
            raise Exception("Cannot allocate capital to " "%s because price is %s as of %s" % (self.name, self._price, self.parent.now))

        # buy/sell
        # determine quantity - must also factor in commission
        # closing out?
        if is_zero(amount + self._value):
            q = -self._position
        else:
            q = amount / (self._price * self.multiplier)
            if self.integer_positions:
                if (self._position > 0) or (is_zero(self._position) and (amount > 0)):
                    # if we're going long or changing long position
                    q = math.floor(q)
                else:
                    # if we're going short or changing short position
                    q = math.ceil(q)

        # if q is 0 nothing to do
        if is_zero(q) or np.isnan(q):
            return

        # unless we are closing out a position (q == -position)
        # we want to ensure that
        #
        # - In the event of a positive amount, this indicates the maximum
        # amount a given security can use up for a purchase. Therefore, if
        # commissions push us above this amount, we cannot buy `q`, and must
        # decrease its value
        #
        # - In the event of a negative amount, we want to 'raise' at least the
        # amount indicated, no less. Therefore, if we have commission, we must
        # sell additional units to fund this requirement. As such, q must once
        # again decrease.
        #
        if not q == -self._position:
            full_outlay, _, _, _ = self.outlay(q)

            # if full outlay > amount, we must decrease the magnitude of `q`
            # this can potentially lead to an infinite loop if the commission
            # per share > price per share. However, we cannot really detect
            # that in advance since the function can be non-linear (say a fn
            # like max(1, abs(q) * 0.01). Nevertheless, we want to avoid these
            # situations.
            # cap the maximum number of iterations to 1e4 and raise exception
            # if we get there
            # if integer positions then we know we are stuck if q doesn't change

            # if integer positions is false then we want full_outlay == amount
            # if integer positions is true then we want to be at the q where
            #   if we bought 1 more then we wouldn't have enough cash
            i = 0
            last_q = q
            last_amount_short = full_outlay - amount
            while not np.isclose(full_outlay, amount, rtol=TOL) and q != 0:
                dq_wout_considering_tx_costs = (full_outlay - amount) / (self._price * self.multiplier)
                q = q - dq_wout_considering_tx_costs

                if self.integer_positions:
                    q = math.floor(q)

                full_outlay, _, _, _ = self.outlay(q)

                # if our q is too low and we have integer positions
                # then we know that the correct quantity is the one  where
                # the outlay of q + 1 < amount. i.e. if we bought one more
                # position then we wouldn't have enough cash
                if self.integer_positions:
                    full_outlay_of_1_more, _, _, _ = self.outlay(q + 1)

                    if full_outlay < amount and full_outlay_of_1_more > amount:
                        break

                # if not integer positions then we should keep going until
                # full_outlay == amount or is close enough

                i = i + 1
                if i > 1e4:
                    raise Exception(
                        "Potentially infinite loop detected. This occurred "
                        "while trying to reduce the amount of shares purchased"
                        " to respect the outlay <= amount rule. This is most "
                        "likely due to a commission function that outputs a "
                        "commission that is greater than the amount of cash "
                        "a short sale can raise."
                    )

                if self.integer_positions and last_q == q:
                    raise Exception(
                        "Newton Method like root search for quantity is stuck!"
                        " q did not change in iterations so it is probably a bug"
                        " but we are not entirely sure it is wrong! Consider "
                        " changing to warning."
                    )
                last_q = q

                if np.abs(full_outlay - amount) > np.abs(last_amount_short):
                    raise Exception(
                        "The difference between what we have raised with q and"
                        " the amount we are trying to raise has gotten bigger since"
                        " last iteration! full_outlay should always be approaching"
                        " amount! There may be a case where the commission fn is"
                        " not smooth"
                    )
                last_amount_short = full_outlay - amount

        self.transact(q, update, False)

    @cy.locals(
        q=cy.double,
        update=cy.bint,
        update_self=cy.bint,
        outlay=cy.double,
        bidoffer=cy.double,
    )
    def transact(self, q, update=True, update_self=True, price=None):
        """
        This transacts the Security. This is the method used to
        buy/sell the security for a given quantity.

        The amount of shares is explicitly provided, a
        commission will be calculated based on the parent's commission fn, and
        any remaining capital will be passed back up  to parent as an
        adjustment.

        Args:
            * amount (float): Amount of adjustment.
            * update (bool): Force update on parent due to transaction proceeds
            * update_self (bool): Check for update on self
            * price (float): Optional price if the transaction happens at a bespoke level
        """
        # will need to update if this has been idle for a while...
        # update if needupdate or if now is stale
        # fetch parent's now since our now is stale
        if update_self and (self._needupdate or self.now != self.parent.now):
            self.update(self.parent.now)

        # if q is 0 nothing to do
        if is_zero(q) or np.isnan(q):
            return

        if price is not None and not self._bidoffer_set:
            raise ValueError('Cannot transact at custom prices when "bidoffer" has ' "not been passed during setup to enable bid-offer tracking.")

        # this security will need an update, even if pos is 0 (for example if
        # we close the positions, value and pos is 0, but still need to do that
        # last update)
        self._needupdate = True

        # adjust position & value
        self._position += q

        # calculate proper adjustment for parent
        # parent passed down amount so we want to pass
        # -outlay back up to parent to adjust for capital
        # used
        full_outlay, outlay, fee, bidoffer = self.outlay(q, p=price)

        # store outlay for future reference
        self._outlay += outlay
        self._bidoffer_paid += bidoffer

        # call parent
        self.parent.adjust(-full_outlay, update=update, flow=False, fee=fee)

    @cy.locals(q=cy.double, p=cy.double)
    def commission(self, q, p):
        """
        Calculates the commission (transaction fee) based on quantity and
        price.  Uses the parent's commission_fn.

        Args:
            * q (float): quantity
            * p (float): price

        """
        return self.parent.commission_fn(q, p)

    @cy.locals(q=cy.double)
    def outlay(self, q, p=None):
        """
        Determines the complete cash outlay (including commission) necessary
        given a quantity q.
        Second returning parameter is a commission itself.

        Args:
            * q (float): quantity
            * p (float): price override
        """
        if p is None:
            fee = self.commission(q, self._price * self.multiplier)
            bidoffer = abs(q) * 0.5 * self._bidoffer * self.multiplier
        else:
            # price override provided: custom transaction
            fee = self.commission(q, p * self.multiplier)
            bidoffer = q * (p - self._price) * self.multiplier

        outlay = q * self._price * self.multiplier + bidoffer

        return outlay + fee, outlay, fee, bidoffer

    def run(self):
        """
        Does nothing - securities have nothing to do on run.
        """
        pass


class Security(SecurityBase):
    """
    A standard security with no special features, and where notional value
    is measured based on market value (notional times price).
    It exists to be able to identify standard securities from nonstandard
    ones via isinstance, i.e. isinstance( sec, Security ) would only return
    True for a vanilla security, whereas SecurityBase would return True for
    all securities.
    """

    pass


class FixedIncomeSecurity(SecurityBase):
    """
    A Fixed Income Security is a security where notional value is
    measured only based on the quantity (par value) of the security.
    Only relevant when using :class:`FixedIncomeStrategy <bt.core.FixedIncomeStrategy>`.
    """

    @cy.locals(coupon=cy.double)
    def update(self, date, data=None, inow=None):
        """
        Update security with a given date and optionally, some data.
        This will update price, value, weight, etc.
        """

        if inow is None:
            if date == 0:
                inow = 0
            else:
                inow = self.data.index.get_loc(date)

        super(FixedIncomeSecurity, self).update(date, data, inow)

        # For fixed income securities (bonds, swaps), notional value is position size, not value!
        self._notl_value = self._position
        self._notl_values.values[inow] = self._notl_value


class CouponPayingSecurity(FixedIncomeSecurity):
    """
    CouponPayingSecurity expands on SecurityBase to handle securities which
    pay (possibly irregular) coupons (or other forms of cash disbursement).
    More generally, this can include instruments with any sort of carry,
    including (potentially asymmetric) holding costs.

    Args:
        * name (str): Security name
        * multiplier (float): security multiplier - typically used for
          derivatives.
        * fixed_income (bool): Flag to control whether notional_value is based
          only on quantity, or on market value (like an equity).
          Defaults to notional weighting for coupon paying instruments.
        * lazy_add (bool): Flag to control whether instrument should be added
          to strategy children lazily, i.e. only when there is a transaction
          on the instrument. This improves performance of strategies which
          transact on a sparse set of children.

    Attributes:
        * SecurityBase attributes
        * coupon (float): Current coupon payment (quantity).
        * holding_cost (float): Current holding cost (quantity).


    Represents a coupon-paying security, where coupon payments adjust
    the capital of the parent. Coupons and costs must be passed in during setup.
    """

    _coupon = cy.declare(cy.double)
    _holding_cost = cy.declare(cy.double)

    @cy.locals(multiplier=cy.double)
    def __init__(self, name, multiplier=1, fixed_income=True, lazy_add=False):
        super(CouponPayingSecurity, self).__init__(name, multiplier)
        self._coupon = 0
        self._holding_cost = 0
        self._fixed_income = fixed_income
        self.lazy_add = lazy_add

    def setup(self, universe, **kwargs):
        """
        Setup Security with universe and coupon data. Speeds up future runs.

        Args:
            * universe (DataFrame): DataFrame of prices with security's name as
              one of the columns.
            * coupons (DataFrame): Manatory DataFrame of coupon/carry amount with
              the same schema as universe.
            * cost_long (DataFrame): Optional DataFrame containing the cost of
              holding a unit long position in the security (i.e. funding).
            * cost_short (DataFrame): Optional DataFrame containing the cost of
              holding a unit short position in the security (i.e. repo).
            * kwargs (dict): Dictionary of additional information needed by
              the strategy. In particular, often takes the form of a DataFrame
              of security level information (i.e. signals, risk, etc).
        """
        super(CouponPayingSecurity, self).setup(universe, **kwargs)

        # Handle coupons
        if "coupons" not in kwargs:
            raise Exception('"coupons" must be passed to setup for a CouponPayingSecurity')

        try:
            self._coupons = kwargs["coupons"][self.name]
        except KeyError:
            self._coupons = None

        if self._coupons is None or not self._coupons.index.equals(universe.index):
            raise ValueError("Index of coupons must match universe data")

        # Handle holding costs
        try:
            self._cost_long = kwargs["cost_long"][self.name]
        except KeyError:
            self._cost_long = None
        try:
            self._cost_short = kwargs["cost_short"][self.name]
        except KeyError:
            self._cost_short = None

        self.data["coupon"] = 0.0
        self.data["holding_cost"] = 0.0
        self._coupon_income = self.data["coupon"]
        self._holding_costs = self.data["holding_cost"]

    @cy.locals(coupon=cy.double, cost=cy.double)
    def update(self, date, data=None, inow=None):
        """
        Update security with a given date and optionally, some data.
        This will update price, value, weight, etc.
        """
        if inow is None:
            if date == 0:
                inow = 0
            else:
                inow = self.data.index.get_loc(date)

        if self._coupons is None:
            raise Exception("coupons have not been set for security %s" % self.name)

        # Standard update
        super(CouponPayingSecurity, self).update(date, data, inow)

        coupon = self._coupons.values[inow]
        # If we were to call self.parent.adjust, then all the child weights would
        # need to be updated. If each security pays a coupon, then this happens for
        # each child. Instead, we store the coupon on self._capital, and it gets
        # swept up as part of the strategy update

        if np.isnan(coupon):
            if is_zero(self._position):
                self._coupon = 0.0
            else:
                raise Exception("Position is open (non-zero) and latest coupon is NaN " "for security %s on %s. Cannot update node value." % (self.name, date))
        else:
            self._coupon = self._position * coupon

        if self._position > 0 and self._cost_long is not None:
            cost = self._cost_long.values[inow]
            self._holding_cost = self._position * cost
        elif self._position < 0 and self._cost_short is not None:
            cost = self._cost_short.values[inow]
            self._holding_cost = -self._position * cost
        else:
            self._holding_cost = 0.0

        self._capital = self._coupon - self._holding_cost
        self._coupon_income.values[inow] = self._coupon
        self._holding_costs.values[inow] = self._holding_cost

    @property
    def coupon(self):
        """
        Current coupon payment (scaled by position)
        """
        if self.root.stale:  # Stale check needed because coupon paid depends on position
            self.root.update(self.root.now, None)
        return self._coupon

    @property
    def coupons(self):
        """
        TimeSeries of coupons paid (scaled by position)
        """
        if self.root.stale:  # Stale check needed because coupon paid depends on position
            self.root.update(self.root.now, None)
        return self._coupon_income.loc[: self.now]

    @property
    def holding_cost(self):
        """
        Current holding cost (scaled by position)
        """
        if self.root.stale:  # Stale check needed because coupon paid depends on position
            self.root.update(self.root.now, None)
        return self._holding_cost

    @property
    def holding_costs(self):
        """
        TimeSeries of coupons paid (scaled by position)
        """
        if self.root.stale:  # Stale check needed because coupon paid depends on position
            self.root.update(self.root.now, None)
        return self._holding_costs.loc[: self.now]


class HedgeSecurity(SecurityBase):
    """
    HedgeSecurity is a SecurityBase where the notional value is set to zero, and thus
    does not count towards the notional value of the strategy. It is intended for use
    in fixed income strategies.

    For example in a corporate bond strategy, the notional value might refer to the size
    of the corporate bond portfolio, and exclude the notional of treasury bonds or interest
    rate swaps used as hedges.
    """

    def update(self, date, data=None, inow=None):
        """
        Update security with a given date and optionally, some data.
        This will update price, value, weight, etc.
        """
        super(HedgeSecurity, self).update(date, data, inow)
        self._notl_value = 0.0
        self._notl_values.values.fill(0.0)


class CouponPayingHedgeSecurity(CouponPayingSecurity):
    """
    CouponPayingHedgeSecurity is a CouponPayingSecurity where the notional value is set to zero, and thus
    does not count towards the notional value of the strategy. It is intended for use
    in fixed income strategies.

    For example in a corporate bond strategy, the notional value might refer to the size
    of the corporate bond portfolio, and exclude the notional of treasury bonds or interest
    rate swaps used as hedges.
    """

    def update(self, date, data=None, inow=None):
        """
        Update security with a given date and optionally, some data.
        This will update price, value, weight, etc.
        """
        super(CouponPayingHedgeSecurity, self).update(date, data, inow)
        self._notl_value = 0.0
        self._notl_values.values.fill(0.0)


class Algo(object):
    """
    Algos are used to modularize strategy logic so that strategy logic becomes
    modular, composable, more testable and less error prone. Basically, the
    Algo should follow the unix philosophy - do one thing well.

    In practice, algos are simply a function that receives one argument, the
    Strategy (referred to as target) and are expected to return a bool.

    When some state preservation is necessary between calls, the Algo
    object can be used (this object). The __call___ method should be
    implemented and logic defined therein to mimic a function call. A
    simple function may also be used if no state preservation is necessary.

    Args:
        * name (str): Algo name

    """

    def __init__(self, name=None):
        self._name = name

    @property
    def name(self):
        """
        Algo name.
        """
        if self._name is None:
            self._name = self.__class__.__name__
        return self._name

    def __call__(self, target):
        raise NotImplementedError("%s not implemented!" % self.name)


class AlgoStack(Algo):
    """
    An AlgoStack derives from Algo runs multiple Algos until a
    failure is encountered.

    The purpose of an AlgoStack is to group a logic set of Algos together. Each
    Algo in the stack is run. Execution stops if one Algo returns False.

    Args:
        * algos (list): List of algos.

    """

    def __init__(self, *algos):
        super(AlgoStack, self).__init__()
        self.algos = algos
        self.check_run_always = any(hasattr(x, "run_always") for x in self.algos)

    def __call__(self, target):
        # normal running mode
        if not self.check_run_always:
            for algo in self.algos:
                if not algo(target):
                    return False
            return True
        # run mode when at least one algo has a run_always attribute
        else:
            # store result in res
            # allows continuation to check for and run
            # algos that have run_always set to True
            res = True
            for algo in self.algos:
                if res:
                    res = algo(target)
                elif hasattr(algo, "run_always"):
                    if algo.run_always:
                        algo(target)
            return res


class Strategy(StrategyBase):
    """
    Strategy expands on the StrategyBase and incorporates Algos.

    Basically, a Strategy is built by passing in a set of algos. These algos
    will be placed in an Algo stack and the run function will call the stack.

    Furthermore, two class attributes are created to pass data between algos.
    perm for permanent data, temp for temporary data.

    Args:
        * name (str): Strategy name
        * algos (list): List of Algos to be passed into an AlgoStack
        * children (dict, list): Children - useful when you want to create
          strategies of strategies
          Children can be any type of Node or str.
          String values correspond to children which will be lazily created
          with that name when needed.
        * parent (Node): The parent Node

    Attributes:
        * stack (AlgoStack): The stack
        * temp (dict): A dict containing temporary data - cleared on each call
          to run. This can be used to pass info to other algos.
        * perm (dict): Permanent data used to pass info from one algo to
          another. Not cleared on each pass.

    """

    def __init__(self, name, algos=None, children=None, parent=None):
        super(Strategy, self).__init__(name, children=children, parent=parent)
        if algos is None:
            algos = []
        self.stack = AlgoStack(*algos)
        self.temp = {}
        self.perm = {}

    def run(self):
        # clear out temp data
        self.temp = {}

        # run algo stack
        self.stack(self)

        # run children
        for c in self._childrenv:
            c.run()


class FixedIncomeStrategy(Strategy):
    """
    FixedIncomeStrategy is an alias for Strategy where the fixed_income flag
    is set to True.

    For this type of strategy:
        - capital allocations are not necessary, and initial capital is not used
        - bankruptcy is disabled (and should be modeled explicitly via an Algo)
        - weights are based off notional_value rather than value
        - strategy price is computed from additive PNL returns
          per unit of current notional_value, with a reference price of PAR.
          :class:`RenormalizedFixedIncomeResult<bt.backtest.RenormalizedFixedIncomeResult>`
          can be used to re-calculate the price-based performance statistics
          using different normalization schemes on total pnl.
        - "transact" assumes the role of "allocate", in order to buy/sell
          children on a weighted notional basis
        - "rebalance" adjusts notionals rather than capital allocations based
          on weights
    """

    def __init__(self, name, algos=None, children=None):
        super(FixedIncomeStrategy, self).__init__(name, algos=algos, children=children)
        self._fixed_income = True



# encoding: utf-8

from typing import List
import streamlit as st

from tgtrader.data import DataGetter
from tgtrader.common import SecurityType
from tgtrader.strategy_config import StrategyConfig

  
class StockDropdownSelectItem:
    def __init__(self, symbol_type:str, code:str, name:str):
        self.symbol_type = symbol_type
        self.code = code
        self.name = name
        
    def __eq__(self, other):
        # 通过 symbol_type 和 code 确定唯一性
        if not isinstance(other, StockDropdownSelectItem):
            return False
        return self.symbol_type == other.symbol_type and self.code == other.code
        
    def __hash__(self):
        # 通过 symbol_type 和 code 确定唯一性
        return hash((self.symbol_type, self.code))
    
def build_stock_dropdown_list(src_page:str, data_getter: DataGetter, strategy_config: StrategyConfig = None) -> list[StockDropdownSelectItem]:
    # 从策略配置中获取默认的证券类型
    default_security_type = None
    if strategy_config is not None and strategy_config.symbols:
        # 获取第一个有标的的证券类型
        for security_type in strategy_config.symbols.keys():
            if security_type == SecurityType.ETF:
                default_security_type = "ETF"
            else:
                default_security_type = "股票"
            break

    # 选择证券类型
    security_type_selectbox = st.selectbox(
        "证券类型", 
        options=["ETF", "股票"],
        index=0 if default_security_type is None else ["ETF", "股票"].index(default_security_type),
        key=f"build_stock_dropdown_list_security_type_selectbox_{src_page}"
    )

    @st.cache_data
    def get_symbols(security_type):
        return data_getter.get_all_symbols(security_type)
  
    # 根据当前选择的类型过滤显示选项
    if security_type_selectbox == "ETF":
        etf_df = get_symbols(SecurityType.ETF)
        etf_options = [f"{row['code']}|{row['name']}|{SecurityType.ETF.value}" for _, row in etf_df.iterrows()]
        display_options = etf_options
    else:
        stock_df = get_symbols(SecurityType.Stocks)
        stock_options = [f"{row['code']}|{row['name']}|{SecurityType.Stocks.value}" for _, row in stock_df.iterrows()]
        display_options = stock_options

    # 如果有策略配置，获取已选择的标的
    default_values = []
    if strategy_config is not None and strategy_config.symbols:
        for security_type, symbols in strategy_config.symbols.items():
            if (security_type == SecurityType.ETF and security_type_selectbox == "ETF") or \
               (security_type == SecurityType.Stocks and security_type_selectbox == "股票"):
                df = get_symbols(security_type)
                for symbol in symbols:
                    row = df[df['code'] == symbol].iloc[0]
                    default_values.append(f"{row['code']}|{row['name']}|{security_type.value}")
    
    # 保持已选择的选项
    symbol_multiselect = st.multiselect(
        f"选择{security_type_selectbox}",
        options=display_options,
        default=default_values,
        format_func=lambda x: '|'.join(x.split('|')[:-1]),
        key=f"build_stock_dropdown_list_symbol_multiselect_{src_page}"
    )

    # 将symbol_multiselect转换为BuildStockDropdownSelectItem列表
    symbol_multiselect = [StockDropdownSelectItem(symbol.split('|')[2], symbol.split('|')[0], symbol.split('|')[1]) for symbol in symbol_multiselect]

    return symbol_multiselect



# encoding: utf-8

import streamlit as st
import pandas as pd
from typing import Dict, List, Any
import plotly.graph_objects as go
from datetime import datetime
import random
from tgtrader.streamlit_pages.utils.common import get_user_name
from tgtrader.dao.t_news_datasource import TNewsDataSource
from tgtrader.utils.db_wrapper import DBWrapper, DBType, DynamicTableWrapper
from tgtrader.dao.t_user_table_meta import UserTableMeta
from tgtrader.utils.defs import USER_TABLE_DB_NAME
from loguru import logger
import json
from st_aggrid import AgGrid, GridOptionsBuilder, GridUpdateMode, DataReturnMode, JsCode


def get_dynamic_table_wrapper():
    table_name = TNewsDataSource.select().where(TNewsDataSource.username == get_user_name()).first().table_name
    if not table_name:
        st.error("请先配置新闻数据源表")
        return
    user_table_meta = UserTableMeta.get_table_meta(get_user_name(), USER_TABLE_DB_NAME, table_name)
    if not user_table_meta:
        st.error("请先配置新闻数据源表")
        return
    db_path = user_table_meta.db_path
    field_config = user_table_meta.columns_info

    dynamic_table_wrapper = DynamicTableWrapper(table_name=table_name, db_path=db_path, db_type=DBType.DUCKDB, field_config=field_config)
    return dynamic_table_wrapper

def get_news_list(limit_cnt=50):
    """
    获取新闻列表，过滤掉国家为空或null的新闻

    Args:
        limit_cnt (int, optional): 获取新闻的数量限制. Defaults to 50.

    Returns:
        list: 过滤后的新闻列表
    """
    table_wrapper = get_dynamic_table_wrapper()
    model = table_wrapper.get_model()
    query = model.select() \
        .where((model.country != "") & model.country.is_null(False)) \
        .limit(limit_cnt) \
        .order_by(model.create_time.desc())

    news_list = list(query.dicts())

    json_columns = ['tags', 'markets', 'related_industries', 'related_indexes', 'related_company']
    
    for news in news_list:
        for column in json_columns:
            if news[column]:
                try:
                    news[column] = json.loads(news[column])
                except json.JSONDecodeError:
                    pass
        
        # 将情绪值从0-1转换为-100到100的范围
        news['sentiment'] = int(float(news['sentiment']) * 100)

    return news_list



def render_tags(tags: List[str]) -> None:
    """
    渲染彩色标签
    
    Args:
        tags: 标签列表
    """
    # CSS样式
    st.markdown("""
        <style>
        .tag-container {
            display: flex;
            flex-wrap: wrap;
            gap: 4px;
            margin: 0;
            padding: 0;
        }
        .tag {
            display: inline-block;
            padding: 2px 10px;
            border-radius: 16px;
            font-size: 14px;
            color: #333;
            white-space: nowrap;
            margin: 2px;
        }
        </style>
    """, unsafe_allow_html=True)
    
    # 淡雅的颜色调色板
    pastel_colors = [
        '#90EE90',  # 浅绿色
        '#87CEFA',  # 浅天蓝色
        '#FFA07A',  # 浅鲑鱼色
        '#20B2AA',  # 浅海洋绿
        '#00CED1',  # 深绿松石色
        '#1E90FF',  # 道奇蓝
        '#FF7F50',  # 珊瑚色
        '#3CB371'   # 中海洋绿
    ]
    
    # 生成标签HTML
    tags_html = '<div class="tag-container">'
    for tag in tags:
        color = random.choice(pastel_colors)
        tags_html += f'<div class="tag" style="background-color: {color};">{tag}</div>'
    tags_html += '</div>'
    
    st.markdown(tags_html, unsafe_allow_html=True)

def render_sentiment_gauge(sentiment: int, title: str) -> None:
    """
    渲染情绪分析仪表盘
    
    Args:
        sentiment: 情绪值 (0-1)
        title: 仪表盘标题
    """
    fig = go.Figure(go.Indicator(
        mode="gauge+number",
        value=sentiment,
        title={
            'text': title, 
            'font': {'size': 16},
        },
        number={'font': {'size': 30}},  # 调整中间数字的大小
        gauge={
            'axis': {'range': [0, 100], 'tickwidth': 1, 'tickcolor': "darkblue"},
            'bar': {'color': "darkblue"},
            'bgcolor': "white",
            'borderwidth': 2,
            'bordercolor': "gray",
            'steps': [
                {'range': [0, 30], 'color': "red"},
                {'range': [30, 70], 'color': "yellow"},
                {'range': [70, 100], 'color': "green"}
            ],
        }
    ))
    fig.update_layout(
        height=125,  # 保持高度
        margin=dict(l=10, r=10, t=40, b=0),  # 增加顶部边距，减少底部边距
        font={'size': 12}
    )
    st.plotly_chart(fig, use_container_width=True)

def render_impact_analysis(data: Dict[str, Any]) -> None:
    """
    渲染影响分析部分
    
    Args:
        data: 新闻数据字典
    """
    # 创建三列布局
    col1, col2, col3 = st.columns(3)
    
    def format_value(value: float) -> str:
        """格式化数值，添加颜色"""
        color = '#FF6B6B' if value >= 0 else '#90EE90'  # 正值红色，负值绿色
        return f'<span style="color: {color}">{value}</span>'
    
    with col1:
        st.subheader("🏭 行业影响")
        for industry in data['related_industries']:
            st.markdown(
                f"**{industry['industry']}**  \n{format_value(float(industry['sentiment']))}",
                unsafe_allow_html=True
            )
    
    with col2:
        st.subheader("📊 指数影响")
        for index in data['related_indexes']:
            st.markdown(
                f"**{index['index']}**  \n{format_value(float(index['sentiment']))}",
                unsafe_allow_html=True
            )
    
    with col3:
        st.subheader("🏢 公司影响")
        for company in data['related_company']:
            st.markdown(
                f"**{company['company']}**  \n{format_value(float(company['sentiment']))}",
                unsafe_allow_html=True
            )

def render_html_content(html_content: str) -> None:
    """
    渲染包含图片的HTML内容，支持多图片显示
    
    Args:
        html_content: 包含HTML标签的文本内容
    """
    current_pos = 0
    while True:
        # 查找下一个图片标签
        img_start = html_content.find('<img', current_pos)
        if img_start == -1:
            # 没有更多图片，显示剩余文本
            remaining_text = html_content[current_pos:].replace('<br />', '\n')
            if remaining_text.strip():
                st.markdown(remaining_text)
            break
        
        # 显示图片前的文本
        text_before_img = html_content[current_pos:img_start].replace('<br />', '\n')
        if text_before_img.strip():
            st.markdown(text_before_img)
        
        # 提取图片URL
        src_start = html_content.find('src="', img_start) + 5
        src_end = html_content.find('"', src_start)
        img_url = html_content[src_start:src_end]
        
        # 显示图片
        st.image(img_url, use_container_width=True)
        
        # 更新位置到图片标签结束
        img_end = html_content.find('>', src_end) + 1
        current_pos = img_end

def render_news_details(data: Dict[str, Any]) -> None:
    """
    渲染新闻详情部分，包括标题、时间、基本信息和图片
    
    Args:
        data: 新闻数据字典，包含标题、时间、国家、市场、标签、描述等信息
    """
    st.caption(f"发布时间: {data['pub_time']}")
    
    # 基本信息
    col1, col2 = st.columns(2)
    with col1:
        st.markdown(f"**国家/地区:** {data['country']}")

        if isinstance(data['markets'], list):
            st.markdown(f"**相关市场:** {', '.join(data['markets'])}")
        else:
            st.markdown(f"**相关市场:** {data['markets']}")
    with col2:
        st.markdown("**标签:**")
        render_tags(data['tags'])
    
    # 渲染整体情绪仪表盘
    st.markdown("### 📈 整体情绪分析")
    render_sentiment_gauge(data['sentiment'], "新闻整体情绪倾向")
    
    # 渲染影响分析
    st.markdown("### 🎯 影响分析")
    render_impact_analysis(data)
    
    # 新闻正文
    st.markdown("### 新闻内容")
    render_html_content(data['description'])

def get_sentiment_color(sentiment: int) -> str:
    """
    根据情绪值返回对应的颜色
    
    Args:
        sentiment: 情绪值 (0-100)
    
    Returns:
        str: 对应的颜色代码
    """
    if sentiment >= 70:
        return '#90EE90'  # 绿色
    elif sentiment >= 30:
        return '#FFD700'  # 黄色
    else:
        return '#FF6B6B'  # 红色

def render_news_list(news_list: list) -> int:
    """
    使用AgGrid渲染新闻列表，返回选中的新闻索引
    
    Args:
        news_list: 新闻列表
    
    Returns:
        int: 选中的新闻索引
    """
    st.header("📋 新闻列表")

    # 初始化 session state
    if 'selected_news_index' not in st.session_state:
        st.session_state.selected_news_index = 0

    # 转换数据为DataFrame格式
    df = pd.DataFrame([{
        '标题': news['title'],
        '时间': news['pub_time'],
        '情绪': news['sentiment'],
        '标签': ', '.join(news['tags'])[:50] + '...' if len(news['tags']) > 50 else ', '.join(news['tags']),
        'index': i  # 添加索引列用于追踪选择
    } for i, news in enumerate(news_list)])

    # 配置情绪列的样式 - 使用中心对齐的水平进度条
    cellrenderer_jscode = JsCode("""
    class SentimentBarRenderer {
        init(params) {
            this.eGui = document.createElement('div');
            this.eGui.style.width = '100%';
            this.eGui.style.height = '20px';
            this.eGui.style.display = 'flex';
            this.eGui.style.alignItems = 'center';
            
            const barContainer = document.createElement('div');
            barContainer.style.width = '80%';
            barContainer.style.height = '12px';
            barContainer.style.backgroundColor = '#f0f0f0';
            barContainer.style.borderRadius = '6px';
            barContainer.style.overflow = 'hidden';
            barContainer.style.position = 'relative';
            
            // 添加中心线
            const centerLine = document.createElement('div');
            centerLine.style.position = 'absolute';
            centerLine.style.left = '50%';
            centerLine.style.width = '1px';
            centerLine.style.height = '100%';
            centerLine.style.backgroundColor = '#999';
            barContainer.appendChild(centerLine);
            
            const bar = document.createElement('div');
            const value = params.value || 0;  // 值范围为-100到100
            
            // 设置bar的位置和宽度
            if (value >= 0) {
                bar.style.left = '50%';
                bar.style.width = (value/2) + '%';  // 除以2是因为50%是总宽度的一半
                bar.style.backgroundColor = '#FF6B6B';  // 正值用红色
            } else {
                bar.style.right = '50%';
                bar.style.width = (Math.abs(value)/2) + '%';  // 除以2是因为50%是总宽度的一半
                bar.style.backgroundColor = '#90EE90';  // 负值用绿色
            }
            
            bar.style.height = '100%';
            bar.style.position = 'absolute';
            bar.style.transition = 'all 0.3s';
            
            const text = document.createElement('span');
            text.style.marginLeft = '5px';
            text.style.fontSize = '12px';
            text.innerText = value;
            
            barContainer.appendChild(bar);
            this.eGui.appendChild(barContainer);
            this.eGui.appendChild(text);
        }
        
        getGui() {
            return this.eGui;
        }
        
        refresh(params) {
            return false;
        }
    }
    """)

    # 配置AgGrid选项
    gb = GridOptionsBuilder.from_dataframe(df)
    gb.configure_selection(selection_mode='single', use_checkbox=False)
    gb.configure_column('标题', minWidth=200)
    gb.configure_column('时间', minWidth=100)
    
    # 配置情绪列
    gb.configure_column('情绪', 
                       minWidth=120,
                       cellRenderer=cellrenderer_jscode)
    
    gb.configure_column('标签', minWidth=150)
    gb.configure_column('index', hide=True)

    # 添加分页设置
    gb.configure_pagination(enabled=True, 
                          paginationAutoPageSize=False, 
                          paginationPageSize=50)

    grid_options = gb.build()

    df = df.sort_values(by='时间', ascending=False)

    # 显示AgGrid表格
    grid_response = AgGrid(
        df,
        gridOptions=grid_options,
        height=1000,
        data_return_mode=DataReturnMode.FILTERED_AND_SORTED,
        update_mode=GridUpdateMode.SELECTION_CHANGED,
        fit_columns_on_grid_load=True,
        theme='streamlit',
        allow_unsafe_jscode=True
    )

    # 处理选择
    selected = grid_response['selected_rows']
    if selected is not None and len(selected) > 0:
        st.session_state.selected_news_index = selected.iloc[0]['index']
    
    return st.session_state.selected_news_index

def run():
    """
    主函数，渲染整个新闻分析页面
    """
    st.title("📊 新闻监控分析")
    
    # 示例数据 - 多条新闻
    sample_news_list = get_news_list(limit_cnt=1000)
    
    # 创建左右两栏布局
    left_col, right_col = st.columns([1, 1])
    
    # 左侧新闻列表
    with left_col:
        selected_index = render_news_list(sample_news_list)
    
    # 右侧新闻详情和分析
    with right_col:
        if selected_index is not None and selected_index < len(sample_news_list):
            selected_news = sample_news_list[selected_index]
            
            # 渲染新闻详情
            render_news_details(selected_news)



from datetime import datetime
import os
import time
from typing import List
from loguru import logger
from peewee import *
import json
import copy
from tgtrader.dao.common import BaseModel, db

class UserTableMeta(BaseModel):
    """
    用户数据元信息.
    """
    
    user = TextField(verbose_name='用户')
    db_name = TextField(verbose_name='数据库名')
    table_name = TextField(verbose_name='表名')
    db_path = TextField(verbose_name='数据库路径')
    columns_info = TextField(verbose_name='列信息')
    version = IntegerField(verbose_name='版本')
    create_time = BigIntegerField(verbose_name='创建时间')
    update_time = BigIntegerField(verbose_name='更新时间')

    class Meta:
        table_name = 't_user_table_meta' 
        primary_key = CompositeKey('user', 'db_name', 'table_name', 'version')
    
    @classmethod
    def init_table(cls):
        # 初始化表
        with db:
            table_exists = cls.table_exists()
            if not table_exists:
                db.create_tables([cls])  # 如果表不存在，创建表

    @classmethod
    def get_all_db_names(cls, user: str):
        data = cls.select(cls.db_name).where(cls.user == user).distinct().execute()
        return [item.db_name for item in data]

    @classmethod
    def get_all_table_names(cls, user: str, db_name: str):
        data = cls.select(cls.table_name).where(cls.user == user, cls.db_name == db_name).distinct().execute()
        return [item.table_name for item in data]
    
    @classmethod
    def get_table_meta(cls, user: str, db_name: str, table_name: str) -> 'UserTableMeta':
        ret = cls.select().where(cls.user == user, cls.db_name == db_name, cls.table_name == table_name).order_by(cls.version.desc()).first()
        if ret:
            ret.columns_info = json.loads(ret.columns_info)
        
        return ret

    @classmethod
    def get_table_columns_info(cls, user: str, db_name: str, table_name: str) -> List[dict]:
        """
        样例：
        [
            {
                "field_name": "code",
                "field_type": "string",
                "description": "code",
                "is_primary_key": true
            }
        ]
        """
        table_info = cls.select().where(cls.user == user, cls.db_name == db_name, cls.table_name == table_name).order_by(cls.version.desc()).first()
        if table_info:
            return json.loads(table_info.columns_info)
        else:
            return []

    @classmethod
    def update_table_meta(cls, user: str, db_name: str, table_name: str, db_path: str, columns_info: List[dict]) -> None:
        """
        更新表的元数据信息，创建新版本记录.

        Args:
            user (str): 用户名
            db_name (str): 数据库名
            table_name (str): 表名
            columns_info (List[dict]): 列信息列表

        Returns:
            None
        """

        columns_info = copy.deepcopy(columns_info)

        for info in columns_info:
            if 'input_field_mapping' in info:
                del info['input_field_mapping']

        # 获取当前最大版本号和最新记录
        latest_record = (cls
            .select()
            .where(
                cls.user == user,
                cls.db_name == db_name,
                cls.table_name == table_name
            )
            .order_by(cls.version.desc())
            .first())
        
        current_max_version = latest_record.version if latest_record else 0
        
        # 创建新版本记录
        current_time = int(time.time())
        new_record = cls.create(
            user=user,
            db_name=db_name,
            table_name=table_name,
            db_path=db_path,
            columns_info=json.dumps(columns_info),
            version=current_max_version + 1,
            create_time=current_time,
            update_time=current_time
        )
        logger.info(f"Created new version {current_max_version + 1} for table {table_name}")

UserTableMeta.init_table()



# encoding: utf-8
import pandas as pd
from tgtrader.common import DataSource
from tgtrader.data_provider.service.akshare_data_service import AkshareDataService
import streamlit as st
from streamlit_ace import st_ace
from .data_meta import build_db_meta_info
from loguru import logger
from tgtrader.dao.t_user_table_meta import UserTableMeta
from tgtrader.utils.db_wrapper import DBWrapper, DBType
from tgtrader.utils.db_path_utils import get_user_data_db_path
from tgtrader.streamlit_pages.utils.common import get_user_name
from tgtrader.dao.t_rss_source import TRssSource
import arrow
from typing import Dict
from tgtrader.dao.t_api_key import TApiKey
from tgtrader.dao.t_llm_template import TLLMTemplate
from tgtrader.utils.defs import USER_TABLE_DB_NAME


def data_source_db_config(node_id: str, src_page: str, node_cfg: dict):
    col1, col2 = st.columns(2)
    with col1:
        options = ["Akshare", "用户自定义数据"]
        default_value = node_cfg.get(
            'data_source', 'Akshare') if node_cfg else 'Akshare'
        default_index = options.index(default_value)
        data_source = st.selectbox("数据源", options,
                                   key=f"data_source_config_{node_id}_{src_page}",
                                   index=default_index)

    placeholder = """
输入：该节点必须为起始节点，不能有其他输入

示例：
select * from t_kdata where date>='2024-01-01'
    """

    sql_query = st_ace(language='sql', theme='dracula',
                       height=300,
                       auto_update=False,
                       placeholder=placeholder,
                       show_gutter=True,
                       key=f"{src_page}_sql_query_{node_id}",
                       value=node_cfg.get('content', '') if node_cfg else ''
                       )

    build_db_meta_info(src_page=f"{src_page}_data_source_config_{node_id}")

    return {
        'type': 'data_source_db',
        'data_source': data_source,
        'content': sql_query
    }


def python_code_config(node_id: str, src_page: str, node_cfg: dict):
    placeholder = """
要定义一个函数，函数名必须为calc，函数参数为前方节点连到该节点的边的名字

输出：
任意类型

示例：
def calc(df):
    df['ma5'] = df['close'].rolling(5).mean()
    df['ma10'] = df['close'].rolling(10).mean()
    df['ma20'] = df['close'].rolling(20).mean()
    return df
"""

    python_code = st_ace(language='python', theme='dracula',
                         height=300, auto_update=False,
                         placeholder=placeholder,
                         show_gutter=True,
                         key=f"{src_page}_python_code_config_{node_id}",
                         value=node_cfg.get('content', '') if node_cfg else ''
                         )

    return {
        'type': 'processor_python_code',
        'content': python_code
    }


def sql_config(node_id: str, src_page: str, node_cfg: dict):
    placeholder = """
支持的输入：
1. SQL语句
2. Python代码

示例：
select * from df where code='000001'
    """

    sql_query = st_ace(language='sql', theme='dracula',
                       height=300, auto_update=False,
                       placeholder=placeholder,
                       show_gutter=True,
                       key=f"{src_page}_sql_config_{node_id}",
                       value=node_cfg.get('content', '') if node_cfg else ''
                       )

    return {
        'type': 'processor_sql',
        'content': sql_query
    }


# 添加字段名映射
FIELD_NAME_MAPPING = {
    'field_name': '字段名',
    'field_type': '类型',
    'description': '描述', 
    'is_primary_key': '是否主键',
    'input_field_mapping': '映射前节点输入字段'
}

def _cn_to_en_field_names(df: pd.DataFrame) -> pd.DataFrame:
    """将DataFrame的中文字段名转换为英文字段名
    
    Args:
        df: 包含中文字段名的DataFrame
        
    Returns:
        包含英文字段名的DataFrame
    """
    reverse_mapping = {v: k for k, v in FIELD_NAME_MAPPING.items()}
    return df.rename(columns=reverse_mapping)

def _en_to_cn_field_names(df: pd.DataFrame) -> pd.DataFrame:
    """将DataFrame的英文字段名转换为中文字段名
    
    Args:
        df: 包含英文字段名的DataFrame
        
    Returns:
        包含中文字段名的DataFrame
    """
    return df.rename(columns=FIELD_NAME_MAPPING)

def _get_empty_field_config():
    """获取空字段配置"""
    return pd.DataFrame({
        "字段名": pd.Series([], dtype='str'),
        "类型": pd.Series([], dtype='str'),
        "描述": pd.Series([], dtype='str'),
        "是否主键": pd.Series([], dtype='bool'),
        "映射前节点输入字段": pd.Series([], dtype='str')
    })

def _create_field_config_editor(node_id: str, src_page: str, is_create_table: bool, field_config: pd.DataFrame):
    """创建字段配置编辑器"""
    if field_config is None or field_config.empty:
        field_config = _get_empty_field_config()

    if is_create_table:
        field_config_df = field_config
        disabled_primary_key = False
    else:
        field_config_df = field_config
        disabled_primary_key = True

    return st.data_editor(
        field_config_df,
        key=f"{src_page}_storage_config_field_config_{'new' if is_create_table else 'existing'}_{node_id}",
        num_rows="dynamic",
        column_config={
            "字段名": st.column_config.TextColumn("字段名", required=True),
            "类型": st.column_config.SelectboxColumn(
                "类型",
                options=["string", "float", "int"],
                required=True
            ),
            "描述": st.column_config.TextColumn("描述", required=True),
            "是否主键": st.column_config.CheckboxColumn(
                "是否主键",
                disabled=disabled_primary_key,
                help="已有表的主键不可修改" if disabled_primary_key else None
            )
        }
    )


def _validate_table_config(table_name: str, field_config_df: pd.DataFrame) -> tuple[bool, str]:
    """验证表配置的合法性"""
    # DuckDB保留关键字列表
    DUCKDB_RESERVED_KEYWORDS = {
        'add', 'all', 'alter', 'and', 'any', 'as', 'asc', 'between', 'by', 'case', 
        'cast', 'check', 'column', 'commit', 'copy', 'create', 'cross', 'current', 
        'default', 'delete', 'desc', 'distinct', 'drop', 'else', 'end', 'escape', 
        'except', 'exists', 'extract', 'false', 'filter', 'following', 'foreign', 
        'from', 'full', 'group', 'having', 'if', 'in', 'inner', 'insert', 'intersect', 
        'into', 'is', 'join', 'left', 'like', 'limit', 'natural', 'not', 'null', 
        'offset', 'on', 'or', 'order', 'outer', 'over', 'primary', 'references', 
        'right', 'rollback', 'select', 'set', 'table', 'then', 'true', 'union', 
        'unique', 'update', 'using', 'values', 'when', 'where', 'window', 'with'
    }

    if not table_name:
        return False, "表名不能为空"

    if field_config_df.empty:
        return False, "字段配置不能为空"

    # 检查必填字段
    required_fields = ["字段名", "类型", "描述"]
    for field in required_fields:
        if field_config_df[field].isnull().any():
            return False, f"{field}不能为空"

    # 检查主键
    if not field_config_df["是否主键"].any():
        return False, "至少需要一个主键字段"

    # 检查字段名唯一性
    if field_config_df["字段名"].duplicated().any():
        return False, "字段名不能重复"
    
    # 检查字段名是否为DuckDB保留关键字
    invalid_fields = [field for field in field_config_df["字段名"] if field.lower() in DUCKDB_RESERVED_KEYWORDS]
    if invalid_fields:
        return False, f"以下字段名是数据库保留关键字，不能使用: {', '.join(invalid_fields)}"
    
    # 检查主键字段是否有映射字段
    primary_key_rows = field_config_df[field_config_df["是否主键"] == True]
    for _, row in primary_key_rows.iterrows():
        if not pd.notna(row.get('映射前节点输入字段', None)):
            return False, "主键字段必须有映射字段"

    return True, ""


def sink_db_config(node_id: str, src_page: str, node_cfg: dict):
    """数据库存储节点配置
    
    Args:
        node_id: 节点ID
        src_page: 源页面
        node_cfg: 节点配置
        
    Returns:
        节点配置字典或None（配置无效时）
    """

    try:
        db_path = get_user_data_db_path(get_user_name())
        db_name = USER_TABLE_DB_NAME
        db_wrapper = DBWrapper(db_path=db_path, db_type=DBType.DUCKDB)

        node_cfg = node_cfg.get('content', {})
        # 初始化配置
        if node_cfg:
            is_create_table = node_cfg.get('is_create_table', False)
            table_name = node_cfg.get('table_name', '')

            # 检查表是否已经存在
            if db_wrapper.is_table_exists(table_name):
                is_create_table = False
        else:
            is_create_table = True
            table_name = ''
            node_cfg = {}

        # 表创建选项
        is_create_table = st.checkbox(
            "是否新建表(保存时就会创建新表)",
            key=f"{src_page}_storage_config_create_table",
            value=is_create_table
        )

        # 表名输入
        col1, col2 = st.columns([1, 5])
        user = get_user_name()
        db_name = 'flow_sinkdb'
        with col1:
            if is_create_table:
                table_name = st.text_input(
                    "表名",
                    placeholder="请输入新建表名",
                    key=f"{src_page}_storage_config_table_name_{node_id}",
                    value=table_name
                )
            else:
                table_names = [''] + UserTableMeta.get_all_table_names(user=user, db_name=db_name)
                
                idx = table_names.index(table_name) if table_name in table_names else 0
                table_name = st.selectbox(
                    "表名",
                    options=table_names,
                    key=f"{src_page}_storage_config_table_name_{node_id}",
                    index=idx
                )
                
            if table_name:
                # 实际的列信息
                real_table_columns = []
                if db_wrapper.is_table_exists(table_name):
                    real_table_columns = db_wrapper.get_columns(table_name)

                real_table_columns_dict = {
                    column.name: {
                        'field_name': column.name,
                        'field_type': column.data_type,
                        'is_primary_key': column.primary_key,
                        'description': '',
                        'input_field_mapping': ''
                    }
                    for column in real_table_columns if column.name not in ['create_time', 'update_time']
                }

                # 读取保存的元信息
                columns_meta_info = UserTableMeta.get_table_columns_info(user=user, db_name=db_name, table_name=table_name)
                columns_meta_info_dict = {
                    column['field_name']: {
                        'field_name': column['field_name'],
                        'field_type': column['field_type'],
                        'is_primary_key': column['is_primary_key'],
                        'description': column['description'],
                        'input_field_mapping': ''
                    }
                    for column in columns_meta_info
                }

                # 传入的配置信息
                field_config = node_cfg.get('field_config', [])
                field_config_dict = {
                    info['field_name']: info
                    for info in field_config
                }
                
                if not real_table_columns and not columns_meta_info:
                    # 情况1：real_table_columns和columns_meta_info都不存在, 使用传入的配置
                    pass
                elif real_table_columns and not columns_meta_info:
                    # 情况2：real_table_columns存在，columns_meta_info不存在，补充新增字段
                    new_columns = set(real_table_columns_dict.keys()) - set(field_config_dict.keys())
                    for column in new_columns:
                        field_config.append(real_table_columns_dict[column])
                elif not real_table_columns and columns_meta_info:
                    # 情况3：real_table_columns不存在，columns_meta_info存在, 使用传入的配置
                    pass
                elif real_table_columns and columns_meta_info:
                    # 情况4：real_table_columns和columns_meta_info都存在，合并三部分信息
                    new_columns = set(real_table_columns_dict.keys()) - set(field_config_dict.keys())
                    # 新增的字段，使用meta_info里的信息进行补充
                    for column in new_columns:
                        if column in columns_meta_info_dict:
                            field_config.append(columns_meta_info_dict[column])
                        else:
                            field_config.append(real_table_columns_dict[column])

                # 将保存的英文字段名转换为中文显示
                field_config = pd.DataFrame(field_config)
                field_config = _en_to_cn_field_names(field_config)
            else:
                field_config = pd.DataFrame()

        # 字段配置
        data_editor_df = _create_field_config_editor(
            node_id, src_page, is_create_table, field_config)

        # 保存按钮
        btn_save = st.button("保存配置", key=f"{src_page}_storage_config_save")

        if btn_save:
            # logger.debug(f"is_create_table: {is_create_table}, table_name: {table_name}, data_editor_df: {data_editor_df}")

            if is_create_table and table_name:
                if db_wrapper.is_table_exists(table_name):
                    st.error(f"表名已存在，请更换表名")
                    ret = None
                else:
                    ret = _save_table_config(table_name, data_editor_df, is_create_table)
            else:
                ret = _save_table_config(table_name, data_editor_df, is_create_table)

        else:
            ret = None

        # 构建元数据信息
        build_db_meta_info(src_page=f"{src_page}_storage_config_{node_id}")
        return ret
    except Exception as e:
        logger.exception(e)
        st.error(f"配置保存失败: {str(e)}")
        return None

def _save_table_config(table_name: str, data_editor_df: pd.DataFrame, is_create_table: bool):
    # 验证配置
    is_valid, error_msg = _validate_table_config(
        table_name, data_editor_df)
    if not is_valid:
        st.error(error_msg)
    else:
        # 保存时转换为英文字段名
        data_editor_df_en = _cn_to_en_field_names(data_editor_df)
        ret = {
            'type': 'sink_db',
            'content': {
                'is_create_table': is_create_table,
                'table_name': table_name,
                'field_config': data_editor_df_en.to_dict(orient='records')
            }
        }
        st.success(f"保存成功，表名：{table_name}")
    
        return ret
    
def data_source_rss_config(node_id: str, src_page: str, node_cfg: dict) -> dict | None:
    """RSS数据源配置组件，使用data_editor实现RSS源的添加和删除.
    
    Args:
        node_id (str): 节点ID
        src_page (str): 源页面
        node_cfg (dict): 节点配置
        
    Returns:
        dict | None: 返回节点配置字典，保存失败时返回None
    """
    # 获取用户名和已配置的RSS源
    username = get_user_name()
    rss_sources = TRssSource.get_rss_sources(username)
    
    # 检查RSS源名称是否唯一
    rss_names = [source.rss_name for source in rss_sources]
    if len(rss_names) != len(set(rss_names)):
        st.error("检测到重复的RSS源名称，请在RSS源管理中修改")
        return None
    
    # 准备RSS源选项和映射
    rss_options = [""] + rss_names
    source_id_map = {source.rss_name: source.id for source in rss_sources}
    
    # 从已有配置中获取选中的RSS源
    selected_data = []
    if node_cfg and 'content' in node_cfg:
        selected_ids = node_cfg['content'].get('selected_rss_ids', [])
        selected_sources = [source for source in rss_sources if source.id in selected_ids]
        selected_data = [{"rss_source": source.rss_name} for source in selected_sources]
    
    # 如果没有选中的源，添加一个空行
    if not selected_data:
        selected_data = [{"rss_source": ""}]
    
    # 创建data editor
    st.write("配置RSS新闻源:")
    edited_data = st.data_editor(
        selected_data,
        column_config={
            "rss_source": st.column_config.SelectboxColumn(
                "RSS源",
                options=rss_options,
                required=True,
                width="large"
            )
        },
        num_rows="dynamic",
        key=f"{src_page}_rss_editor_{node_id}",
        use_container_width=True
    )

    # logger.debug(f"edited_data: {edited_data}")
    
    # 保存按钮
    if st.button("保存配置", key=f"{src_page}_rss_config_save_{node_id}"):
        # 检查是否有重复选择
        selected_sources = [row.get("rss_source") for row in edited_data if row.get("rss_source")]
        if len(selected_sources) != len(set(selected_sources)):
            st.error("RSS源不能重复选择")
            return None
            
        # 收集选中的RSS源ID
        new_selected_ids = []
        for row in edited_data:
            rss_name = row.get("rss_source")
            if rss_name and rss_name in source_id_map:
                new_selected_ids.append(source_id_map[rss_name])
        
        if not new_selected_ids:
            st.error("请至少选择一个RSS源")
            return None
        
        # 收集选中的RSS源的详细信息
        rss_info = []
        for source in rss_sources:
            if source.id in new_selected_ids:
                rss_info.append({
                    "id": source.id,
                    "url": source.rss_url
                })
        
        config = {
            'type': 'data_source_rss',
            'content': {
                'selected_rss_ids': new_selected_ids,
                'rss_info': rss_info
            }
        }
        st.success("保存成功")
        return config
    
    return None
    
def processor_llm_config(node_id: str, src_page: str, node_cfg: dict) -> dict | None:
    """LLM处理节点配置组件
    
    Args:
        node_id: 节点ID
        src_page: 源页面
        node_cfg: 节点配置
        
    Returns:
        节点配置字典或None（配置无效时）
    """
    # logger.debug(f"node_cfg: {node_cfg}")
    # 获取用户名
    username = get_user_name()
    if not username:
        st.error("请先登录")
        return None
        
    # 获取用户配置的API Keys
    api_keys = TApiKey.get_api_keys(username=username, hide_middle=True)
    if not api_keys:
        st.error("请先在设置页面配置API Key")
        return None
    
    model_name_id_key = {key.api_key_name: key.id for key in api_keys}
    model_id_name_key = {key.id: key.api_key_name for key in api_keys}
        
    # 准备模型选项
    model_options = list(set(key.api_key_name for key in api_keys))
    
    # 从已有配置中获取选中的模型和模板
    selected_model_id = None
    selected_template = None
    prompt_template = ""
    if node_cfg and 'content' in node_cfg:
        selected_model_id = node_cfg['content'].get('model_id', '')
        selected_template = node_cfg['content'].get('template_name', '')
        prompt_template = node_cfg['content'].get('prompt_template', '')
    
    # 模型选择下拉框
    if selected_model_id:
        selected_model = model_id_name_key[selected_model_id]
    else:
        selected_model = None
    
    selected_model = st.selectbox(
        "选择模型",
        options=model_options,
        index=model_options.index(selected_model) if selected_model and selected_model in model_options else 0,
        key=f"{src_page}_llm_model_{node_id}"
    )
    
    # 获取所有模板
    templates = TLLMTemplate.get_all_template()
    template_options = [name for name, template in templates.items()]
    # 存一个dict
    template_dict = {name: template for name, template in templates.items()}
    
    # 模板选择下拉框
    selected_template = st.selectbox(
        "选择模板",
        options=template_options,
        index=template_options.index(selected_template) if selected_template in template_options else 0,
        key=f"{src_page}_llm_template_{node_id}"    
    )

    # 如果选择了模板，更新prompt_template
    if selected_template:
        if not prompt_template: 
            prompt_template = template_dict[selected_template]
    else:
        prompt_template = ''

    value = st.text_area(
        "提示词模板",
        key=f"{src_page}_llm_prompt_{node_id}",
        value=prompt_template,
        height=300
    )

    
    btn_save = st.button("保存配置", key=f"{src_page}_llm_config_save_{node_id}")
    
    if btn_save:
        config = {
                'type': 'processor_llm',
                'content': {
                    'model_id': model_name_id_key[selected_model],
                    'template_name': selected_template,
                    'prompt_template': value,
                }
            }
        
        return config
    
    return None


# encoding: utf-8
from peewee import *
from tqdm import tqdm
from loguru import logger
from tgtrader.common import DataSource
from tgtrader.data_provider.dao.akshare.common import main_db
from tgtrader.data_provider.dao.models.common import register_model
from tgtrader.data_provider.dao.models.t_kdata_model import T_KData_Model


@register_model(DataSource.Akshare, 't_kdata')
class T_KData(T_KData_Model):
    class Meta:
        database = main_db

    @classmethod
    def init_table(cls):
        # 初始化表
        with main_db:
            table_exists = T_KData.table_exists()
            if not table_exists:
                main_db.create_tables([T_KData])  # 如果表不存在，创建表

    @classmethod
    def batch_insert_many(cls, data_list: list, batch_size: int = 1000) -> int:
        """批量插入数据，自动分批、显示进度

        Args:
            data_list: 要插入的数据列表
            batch_size: 每批数据量，默认1000条

        Returns:
            int: 成功插入的记录数
        """
        if not data_list:
            logger.warning("No data to insert")
            return 0

        total_count = 0
        
        try:
            with main_db:
                # 使用tqdm显示进度
                for i in tqdm(range(0, len(data_list), batch_size), 
                            desc="Inserting data"):
                    batch = data_list[i:i + batch_size]
                    rows = cls.insert_many(batch).on_conflict(
                        conflict_target=[cls.code, cls.date, cls.source],
                        action='UPDATE',
                        update={
                            cls.open: cls.open,
                            cls.high: cls.high,
                            cls.low: cls.low,
                            cls.close: cls.close,
                            cls.volume: cls.volume,
                            cls.adjust_type: cls.adjust_type,
                            cls.update_time: cls.update_time
                        }
                    ).execute()
                    total_count += len(batch)
                    
            logger.info(f"Successfully inserted/updated {total_count} records")
            return total_count
            
        except Exception as e:
            logger.error(f"Error in batch insert: {str(e)}")
            raise



import streamlit as st
import pandas as pd
import altair as alt  # Streamlit recommends Altair for interactive charts
from typing import Dict, Any, Optional

from tgtrader.strategy import PerformanceStats, StrategyDef

def plot_returns_chart(prices: pd.DataFrame, backtest_end_date: Optional[str] = None):
    """绘制策略收益率曲线，若提供backtest_end_date，则分割回测与非回测部分并加粗非回测部分的线条"""
    
    print(prices.iloc[0])
    print(prices.iloc[-1])
    print(backtest_end_date)

    # 确保日期列为datetime类型
    plot_df = prices.reset_index()
    plot_df.columns = ['date', 'returns']
    plot_df['date'] = pd.to_datetime(plot_df['date'])
    
    if backtest_end_date:
        backtest_end = pd.to_datetime(backtest_end_date)
        # 分割数据为回测和非回测部分
        backtest_df = plot_df[plot_df['date'] <= backtest_end]
        non_backtest_df = plot_df[plot_df['date'] > backtest_end]
        
        # 最小和最大收益率，用于y轴范围
        min_return = plot_df['returns'].min()
        max_return = plot_df['returns'].max()
        y_padding = (max_return - min_return) * 0.05
        
        # 创建回测部分的线
        backtest_line = alt.Chart(backtest_df).mark_line(color='blue').encode(
            x=alt.X('date:T', title='日期'),
            y=alt.Y('returns:Q', 
                    title='累计收益率',
                    scale=alt.Scale(
                        domain=[min_return - y_padding, max_return + y_padding]
                    ))
        )
        
        # 创建非回测部分的线，线条更粗
        non_backtest_line = alt.Chart(non_backtest_df).mark_line(color='red', strokeWidth=3).encode(
            x='date:T',
            y='returns:Q'
        )
        
        # 创建分割竖线
        rule = alt.Chart(pd.DataFrame({'date': [backtest_end]})).mark_rule(color='green').encode(
            x='date:T'
        )
        
        # 合并所有图层
        chart = (backtest_line + non_backtest_line + rule).properties(
            height=500
        )
    else:
        # 如果没有提供backtest_end_date，则绘制普通的线图
        min_return = plot_df['returns'].min()
        max_return = plot_df['returns'].max()
        y_padding = (max_return - min_return) * 0.05
        
        chart = alt.Chart(plot_df).mark_line().encode(
            x=alt.X('date:T', title='日期'),
            y=alt.Y('returns:Q', 
                    title='累计收益率',
                    scale=alt.Scale(
                        domain=[min_return - y_padding, max_return + y_padding]
                    )),
            tooltip=['date:T', 'returns:Q']
        ).properties(
            height=500
        )
    
    return chart

def format_percentage(value: float) -> str:
    """将小数转换为百分比格式"""
    return f"{value * 100:.2f}%"

def format_ratio(value: float) -> str:
    """格式化比率数值"""
    return f"{value:.2f}"

def display_statistics(stats: PerformanceStats):
    """显示回测统计信息"""
    
    # 收益指标
    returns_metrics = {
        "总收益率": format_percentage(stats.total_return),
        "年化收益率": format_percentage(stats.cagr), 
        "最大回撤": format_percentage(stats.max_drawdown),
        "平均回撤": format_percentage(stats.avg_drawdown),
        "平均回撤天数": str(stats.avg_drawdown_days),
        "Calmar比率": format_ratio(stats.calmar)
    }
    
    # 风险指标
    risk_metrics = {
        "日度夏普比率": format_ratio(stats.daily_sharpe),
        "日度索提诺比率": format_ratio(stats.daily_sortino),
        "日度波动率": format_percentage(stats.daily_vol),
        "月度夏普比率": format_ratio(stats.monthly_sharpe),
        "月度索提诺比率": format_ratio(stats.monthly_sortino),
        "月度波动率": format_percentage(stats.monthly_vol)
    }
    
    # 创建两列布局
    col1, col2 = st.columns(2)
    
    # 显示收益指标
    with col1:
        st.subheader("收益指标")
        returns_df = pd.DataFrame.from_dict(returns_metrics, orient='index', columns=['值'])
        st.dataframe(returns_df, use_container_width=True)
    
    # 显示风险指标
    with col2:
        st.subheader("风险指标")
        risk_df = pd.DataFrame.from_dict(risk_metrics, orient='index', columns=['值'])
        st.dataframe(risk_df, use_container_width=True)

def display_backtest_results(strategy: StrategyDef, backtest_end_date: Optional[str] = None):
    """显示回测结果的主函数"""
    st.header("回测结果")
    
    # 绘制收益率曲线
    chart = plot_returns_chart(strategy.get_prices(), backtest_end_date)
    st.altair_chart(chart, use_container_width=True)
    
    display_statistics(strategy.performance_stats())



import streamlit as st
from tgtrader.service.account_service import AccountService

def login_page():
    """Display the login page"""
    st.title('登录系统')
    
    # Add debug information
    # st.write("Debug: Session State:", st.session_state)
    
    # 创建登录表单
    login_form = st.form(key="login_form")
    with login_form:
        username = st.text_input('用户名', key='username_input')
        password = st.text_input('密码', type='password', key='password_input')
        submit_button = st.form_submit_button('登录')
    
    # Form submission handling
    if submit_button:
        user_info = AccountService.verify_user(username, password)
        if user_info:
            st.success('登录成功！')
            st.session_state['logged_in'] = True
            st.session_state['user_info'] = user_info
            st.rerun()
        else:
            st.error('用户名或密码错误！')

def run():
    if 'logged_in' not in st.session_state:
        st.session_state['logged_in'] = False
        st.session_state['user_info'] = None
    
    if not st.session_state['logged_in']:
        login_page()



# encoding: utf-8

from typing import Callable, Dict, Any, List
import pandas as pd
from peewee import Model
from tgtrader.flow.flow import FlowNode, register_node
import os
from tgtrader.utils.db_wrapper import DBWrapper, DBType
from loguru import logger
from tgtrader.utils.db_path_utils import get_user_data_db_path
from tgtrader.dao.t_user_table_meta import UserTableMeta
from tgtrader.utils.defs import USER_TABLE_DB_NAME
import json


# DuckDB 保留关键字列表
DUCKDB_RESERVED_KEYWORDS = {
    'add', 'all', 'alter', 'and', 'any', 'as', 'asc', 'between', 'by', 'case', 
    'cast', 'check', 'column', 'commit', 'copy', 'create', 'cross', 'current', 
    'default', 'delete', 'desc', 'distinct', 'drop', 'else', 'end', 'escape', 
    'except', 'exists', 'extract', 'false', 'filter', 'following', 'foreign', 
    'from', 'full', 'group', 'having', 'if', 'in', 'inner', 'insert', 'intersect', 
    'into', 'is', 'join', 'left', 'like', 'limit', 'natural', 'not', 'null', 
    'offset', 'on', 'or', 'order', 'outer', 'over', 'primary', 'references', 
    'right', 'rollback', 'select', 'set', 'table', 'then', 'true', 'union', 
    'unique', 'update', 'using', 'values', 'when', 'where', 'window', 'with'
}


@register_node("存储(DB)")
class SinkDBNode(FlowNode):
    @staticmethod
    def _validate_field_names(field_config: List[Dict[str, Any]]) -> None:
        """验证字段名是否使用了DuckDB保留关键字.
        
        Args:
            field_config: 字段配置列表
            
        Raises:
            ValueError: 当字段名使用了保留关键字时抛出
        """
        invalid_fields = []
        for field in field_config:
            field_name = field.get('field_name', '').lower()
            if field_name in DUCKDB_RESERVED_KEYWORDS:
                invalid_fields.append(field_name)
        
        if invalid_fields:
            raise ValueError(f"以下字段名是DuckDB保留关键字，不能使用: {', '.join(invalid_fields)}")

    def execute(self, input_data: dict, process_callback: Callable = None) -> None:
        """执行数据库写入操作.
        
        Args:
            input_data: 包含DataFrame的输入数据字典
            process_callback: 处理回调函数
            
        Returns:
            None
        """
        
        last_progress = -1
        def percent_callback(progress: int):
            nonlocal last_progress
            if process_callback and progress != last_progress:
                process_callback(f"【节点: {self.node_label}】进度: {progress}%", message_type="info")
                last_progress = progress

        try:
            content = self.config.get('content', None)
            if not content:
                if process_callback:
                    process_callback(f"【节点: {self.node_label}】config里未提供content", message_type="error")
                raise ValueError("config里未提供content")
            
            is_create_table = content['is_create_table']
            table_name = content['table_name']
            field_config = content['field_config']

            if any([is_create_table is None, table_name is None, field_config is None]):
                if process_callback:
                    process_callback(f"【节点: {self.node_label}】config里未提供is_create_table, table_name, field_config", message_type="error")
                raise ValueError("config里未提供is_create_table, table_name, field_config")
            
            # 验证字段名
            try:
                self._validate_field_names(field_config)
            except ValueError as e:
                if process_callback:
                    process_callback(f"【节点: {self.node_label}】字段名验证失败: {str(e)}", message_type="error")
                raise
            
            db_path = get_user_data_db_path(self.user)
            db_name = USER_TABLE_DB_NAME

            # 添加初始进度回调
            if process_callback:
                process_callback(f"【节点: {self.node_label}】准备连接数据库", message_type="info")

            db_wrapper = DBWrapper(db_path=db_path, db_type=DBType.DUCKDB)

            if is_create_table:
                # 检查表是否存在
                if not db_wrapper.is_table_exists(table_name):
                    if process_callback:
                        process_callback(f"【节点: {self.node_label}】创建新表", message_type="info")
                    db_wrapper.create_table(table_name, field_config)
                    
            else:
                # 检查表是否存在
                if not db_wrapper.is_table_exists(table_name):
                    if process_callback:
                        process_callback(f"【节点: {self.node_label}】表{table_name}不存在", message_type="error")
                    raise Exception(f"表{table_name}不存在")

            # 对比field_config和meta_info
            if process_callback:
                process_callback(f"【节点: {self.node_label}】检查表结构", message_type="info")

            meta_info = UserTableMeta.get_table_columns_info(self.user, USER_TABLE_DB_NAME, table_name)
            if self.compare_field_config(meta_info, field_config):
                # 是否有新增字段
                old_columns = db_wrapper.get_columns(table_name)
                old_columns = [column.name for column in old_columns]

                add_columns = []
                for new_field in field_config:
                    if new_field['field_name'] not in old_columns:
                        add_columns.append(new_field)
                
                if add_columns:
                    if process_callback:
                        process_callback(f"【节点: {self.node_label}】发现新字段: {add_columns}", message_type="warning")
                    logger.warning(f"新增字段: {add_columns}")
                    db_wrapper.add_column(table_name, add_columns)

                UserTableMeta.update_table_meta(self.user, db_name, table_name, db_path, field_config)

            # data进行处理
            columns_mapping = dict()
            for field in field_config:
                input_field_mapping = field.get('input_field_mapping', None)
                field_name = field.get('field_name', None)
                if input_field_mapping:
                    columns_mapping[input_field_mapping] = field_name

            # 导入数据
            total_records = 0
            total_dfs = len(input_data)
            for idx, (_, df) in enumerate(input_data.items(), 1):

                if df is None or df.empty:
                    if process_callback:
                        process_callback(f"【节点: {self.node_label}】数据集 {idx}/{total_dfs} 为空，跳过", message_type="warning")
                        continue
                    
                if process_callback:
                    process_callback(f"【节点: {self.node_label}】处理数据集 {idx}/{total_dfs}", message_type="info")

                reserved_columns = columns_mapping.keys()

                # df的列，与reserved_columns求交集
                df_columns = df.columns
                reserved_columns = [col for col in reserved_columns if col in df_columns]

                df = df[reserved_columns]
                df.rename(columns=columns_mapping, inplace=True)

                if 'create_time' not in df.columns:
                    df['create_time'] = pd.Timestamp.now().timestamp() * 1000
                if 'update_time' not in df.columns:
                    df['update_time'] = pd.Timestamp.now().timestamp() * 1000

                if process_callback:
                    process_callback(f"【节点: {self.node_label}】开始写入数据集 {idx}", message_type="info")

                db_wrapper.insert_data(
                    table_name, df, progress_percent_callback=percent_callback)
                
                total_records += len(df)

            if process_callback:
                process_callback(f"【节点: {self.node_label}】数据写入完成", message_type="success")
                
        except Exception as e:
            if process_callback:
                process_callback(f"【节点: {self.node_label}】数据库写入失败: {str(e)}", message_type="error")
            raise

    def compare_field_config(self, arr1, arr2):
        # 转换为忽略input_field_mapping的字典
        to_dict = lambda arr: {x["field_name"]: {k:v for k,v in x.items() if k != "input_field_mapping"} for x in arr}
        d1, d2 = to_dict(arr1), to_dict(arr2)
        
        # 找出所有差异
        added = [{"field":k, "type":"added"} for k in d2.keys() - d1.keys()]
        removed = [{"field":k, "type":"removed"} for k in d1.keys() - d2.keys()]
        
        # 找出修改项
        modified = []
        for k in d1.keys() & d2.keys():
            diff = {key: (d1[k][key], d2[k][key]) for key in d1[k] if d1[k][key] != d2[k].get(key)}
            if diff:
                modified.append({"field":k, "type":"modified", "details":diff})
        
        return added + removed + modified


# encoding: utf-8
from loguru import logger
from peewee import *
from datetime import datetime
from tgtrader.dao.common import BaseModel, db

class UserStrategy(BaseModel):
    # 主键自增id
    id = AutoField()
    # 用户id
    user_id = IntegerField()
    # 策略配置JSON
    strategy = TextField()
    # 创建时间
    create_time = BigIntegerField(default=lambda: int(datetime.now().timestamp()))
    # 更新时间
    update_time = BigIntegerField(default=lambda: int(datetime.now().timestamp()))

    class Meta:
        table_name = 't_user_strategy'

    def save(self, *args, **kwargs):
        # 更新时间戳
        self.update_time = int(datetime.now().timestamp())
        return super(UserStrategy, self).save(*args, **kwargs)

    @classmethod
    def init_table(cls):
        # 初始化表
        with db:
            if not UserStrategy.table_exists():
                db.create_tables([UserStrategy])  # 如果表不存在，创建表

    @classmethod
    def get_user_strategies(cls, user_id: int):
        """获取用户的所有策略，按更新时间倒序排列"""
        with db:
            return list(cls.select().where(cls.user_id == user_id).order_by(cls.update_time.desc()))
        
    @classmethod
    def get_strategy(cls, user_id: int, strategy_id: int):
        """获取策略"""
        with db:
            return cls.get(cls.id == strategy_id, cls.user_id == user_id)

    @classmethod
    def create_strategy(cls, user_id: int, strategy: str):
        """创建新的策略"""
        with db:
            return cls.create(
                user_id=user_id,
                strategy=strategy
            )

    @classmethod
    def update_strategy(cls, user_id: int, strategy_id: int, strategy: str):
        """更新策略"""
        with db:
            query = cls.update(
                strategy=strategy,
                update_time=int(datetime.now().timestamp())
            ).where(cls.id == strategy_id, cls.user_id == user_id)
            return query.execute()

    @classmethod
    def delete_strategy(cls, user_id: int, strategy_id: int):
        """删除策略"""
        with db:
            return cls.delete().where(cls.id == strategy_id, cls.user_id == user_id).execute()


UserStrategy.init_table()



# encoding: utf-8
from datetime import datetime
from tgtrader.dao.common import BaseModel, db
from peewee import AutoField, IntegerField, BigIntegerField, TextField, CompositeKey


class TNewsDataSource(BaseModel):
    # 主键自增id
    id = AutoField()
    # 用户名
    username = TextField()
    # 表名
    table_name = TextField()
    # 创建时间
    create_time = BigIntegerField(
        default=lambda: int(datetime.now().timestamp()))
    # 更新时间
    update_time = BigIntegerField(
        default=lambda: int(datetime.now().timestamp()))

    class Meta:
        table_name = 't_news_datasource'
        indexes = (
            (('username', 'table_name'), True),
        )

    @classmethod
    def init_table(cls):
        with db:
            if not TNewsDataSource.table_exists():
                db.create_tables([TNewsDataSource])

TNewsDataSource.init_table()



# encoding: utf-8

import time
from tgtrader.dao.t_flow import FlowCfg
from tgtrader.dao.t_flow_node import FlowNodeCfg
import json
from typing import Callable, List
from tgtrader.flow.flow import Flow
from loguru import logger

class FlowConfigService:
    def __init__(self):
        pass
    
    @classmethod
    def save_node(cls, flow_id, node_id, node_type, node_cfg):
        """
        根据flow_id和node_id，查询是否有is_draft=1的记录，如果有，则更新，否则插入
        如果是插入，并且之前已经有相同flow_id和node_id的记录，则version+1
        """

        # 查询是否有草稿状态的记录
        existing_draft = FlowNodeCfg.select().where(
            (FlowNodeCfg.flow_id == flow_id) & 
            (FlowNodeCfg.node_id == node_id) & 
            (FlowNodeCfg.is_draft == 1)
        ).first()

        current_timestamp = int(time.time() * 1000)

        if existing_draft:
            # 如果有草稿状态的记录，直接更新
            existing_draft.node_type = node_type.value
            existing_draft.node_cfg = json.dumps(node_cfg)
            existing_draft.update_time = current_timestamp
            existing_draft.save()
        else:
            # 查询该flow_id和node_id的最大版本号
            max_version_record = FlowNodeCfg.select(
                FlowNodeCfg.version
            ).where(
                (FlowNodeCfg.flow_id == flow_id) & 
                (FlowNodeCfg.node_id == node_id)
            ).order_by(
                FlowNodeCfg.version.desc()
            ).first()

            # 确定新的版本号
            new_version = max_version_record.version + 1 if max_version_record else 1

            # 创建新的草稿记录
            FlowNodeCfg.create(
                flow_id=flow_id,
                node_id=node_id,
                node_type=node_type.value,
                node_cfg=json.dumps(node_cfg),
                version=new_version,
                is_draft=1,
                create_time=current_timestamp,
                update_time=current_timestamp
            )

    @classmethod
    def save_flow(cls, username, flow_id, flow_type, flow_name, node_list, edge_list):
        """
        新增：
            1. t_flow表新增一条记录
            2. t_flow的node_list里所有记录，将is_draft=1的记录，更新为is_draft=0
        修改：
            1. t_flow表，更新flow_id=flow_id的记录
            2. t_flow的node_list里所有记录，将is_draft=1的记录，更新为is_draft=0
        """
        current_timestamp = int(time.time() * 1000)

        # 检查是否已存在相同flow_id的记录
        existing_flow = FlowCfg.get_or_none(FlowCfg.flow_id == flow_id)

        if existing_flow:
            # 更新已存在的流程记录
            existing_flow.flow_type = flow_type
            existing_flow.flow_name = flow_name
            existing_flow.node_list = json.dumps(node_list)
            existing_flow.edge_list = json.dumps(edge_list)
            existing_flow.update_time = current_timestamp
            existing_flow.username = username
            existing_flow.save()
        else:
            # 检查是否存在相同flow_name的记录
            existing_flow_name = FlowCfg.select().where(
                (FlowCfg.username == username) & 
                (FlowCfg.flow_name == flow_name)
            ).first()

            if existing_flow_name:
                raise Exception(f"流程名称已存在，请修改流程名称")

            # 创建新的流程记录
            FlowCfg.create(
                username=username,
                flow_id=flow_id,
                flow_type=flow_type,
                flow_name=flow_name,
                node_list=json.dumps(node_list),
                edge_list=json.dumps(edge_list),
                create_time=current_timestamp,
                update_time=current_timestamp
            )

        # 更新节点记录的草稿状态
        for node in node_list:
            node_id = node['id']
            draft_nodes = FlowNodeCfg.select().where(
                (FlowNodeCfg.flow_id == flow_id) & 
                (FlowNodeCfg.node_id == node_id) & 
                (FlowNodeCfg.is_draft == 1)
            )
            
            for node in draft_nodes:
                node.is_draft = 0
                node.save()

    @classmethod
    def get_node_info(cls, flow_id, node_id) -> FlowNodeCfg:
        node_info = FlowNodeCfg.select().where(
            (FlowNodeCfg.flow_id == flow_id) &
            (FlowNodeCfg.node_id == node_id)
        ).order_by(FlowNodeCfg.version.desc()).first()
        return node_info
    
    @classmethod
    def get_flow_info(cls, flow_id) -> FlowCfg:
        flow_info = FlowCfg.get_or_none(FlowCfg.flow_id == flow_id)
        if not flow_info:
            return None
        flow_info.node_list = json.loads(flow_info.node_list)
        flow_info.edge_list = json.loads(flow_info.edge_list)
        return flow_info

    @classmethod
    def get_all_user_flow_info(cls, username, flow_type):
        flow_info = FlowCfg.select().where(
            (FlowCfg.username == username) & 
            (FlowCfg.flow_type == flow_type)
        ).order_by(FlowCfg.create_time.desc())
        return flow_info

    @classmethod
    def get_user_flow_info_by_flow_name(cls, username, flow_name):
        if not flow_name:
            return None

        flow_info = FlowCfg.select().where(
            (FlowCfg.username == username) & 
            (FlowCfg.flow_name == flow_name)
        ).first()
        return flow_info

    @classmethod
    def delete_node_draft(cls, user_name, flow_id):
        """删除用户某个flow下所有is_draft=1的节点配置
        
        Args:
            user_name (str): 用户名
            flow_id (str): 流程id
        """
        # 获取flow信息
        flow_info = cls.get_flow_info(flow_id)
        if not flow_info or flow_info.username != user_name:
            return
        
        # 删除所有草稿状态的节点配置
        draft_nodes = FlowNodeCfg.delete().where(
            (FlowNodeCfg.flow_id == flow_id) &
            (FlowNodeCfg.is_draft == 1)
        ).execute()


    @staticmethod
    def __default_run_flow_callback(message, message_type):
        if message_type == "info":
            logger.info(message)
        elif message_type == "warning":
            logger.warning(message)
        elif message_type == "error":
            logger.error(message)
        else:
            logger.info(message)

    @classmethod
    def run_flow(cls, user: str, flow_id, info_callback: Callable=__default_run_flow_callback):
        """
        info_callback: 
            Args: 
                message: 消息
                message_type: 消息类型，可选值：info, warning, error
        
        process_callback: 
            Args: 
                data: 处理后的数据
        """
        if info_callback:
            info_callback("开始运行流程", message_type="info")

        flow_info = FlowConfigService.get_flow_info(flow_id)

        if not flow_info:
            info_callback("未找到flow信息, 请先保存流程", message_type="error")
            return

        try:
            flow = Flow(flow_id, user=user)

            node_list = cls.__convert_node_list(flow_id, flow_info.node_list)
            edge_list = cls.__convert_edge_list(flow_info.edge_list)
            flow.build_flow(node_list, edge_list)

            flow.execute_flow(info_callback)
        except Exception as e:
            info_callback(f"流程构建失败: {e}", message_type="error")
            logger.exception(e)
            return

    @classmethod
    def __convert_node_list(cls, flow_id, node_list: List[dict]) -> list:
        ret_list = []
        for node in node_list:
            node_id = node['id']
            node_label = node['data']['content']

            node_info = FlowNodeCfg.select().where(
                (FlowNodeCfg.flow_id == flow_id) &
                (FlowNodeCfg.node_id == node_id)
            ).order_by(FlowNodeCfg.version.desc()).first()

            if not node_info:
                raise Exception(f"未找到节点{node_id}的配置信息")
            
            node_type = node_info.node_type
            node_cfg = node_info.node_cfg

            ret_list.append({
                "id": node_id,
                "node_label": node_label,
                "node_type": node_type,
                "config": node_cfg
            })

        return ret_list

    @classmethod
    def __convert_edge_list(cls, edge_list: List[dict]) -> list:
        ret_list = []
        for edge in edge_list:
            ret_list.append({
                "source": edge['source'],
                "target": edge['target'],
                "edge_name": edge['label']
            })
        return ret_list




# 欢迎页面

import streamlit as st

def run():
    st.markdown("""
# tgtrader天工量化投研分析客户端

- 提供开箱即用的分析工具
- 对于小白，可以直接使用可视化分析工具，零代码开启量化分析
- 对于有一定经验的开发者，结合使用sdk开发，更加灵活

## 微信公众号： 天工量化

**关注即可获取**：

1. tgtrader的最佳实践：如何用好tgtrader以提高投研效率
2. 研报复现：各大券商研报复现，源码公开
3. 策略分享：基于tg量化工具集，实现各类策略
4. 实盘跟踪：已上线的实盘策略持续跟进

## 源码地址

github: https://github.com/smartian1/tgtrader

gitee: https://gitee.com/smartian123/tgtrader   

## 安装

```sh
pip install tgtrader

如果网络不通，可以使用国内镜像源
pip install tgtrader -i https://mirrors.aliyun.com/pypi/simple/
```



## 更新日志

**v1.2.0**

- 数据流支持使用AI大模型解读财经新闻
- 支持任务调度，定时执行数据加工流，用于实时解析财经新闻



**v1.1.0**

- 支持画布拖拽构建数据加工流程，无需写代码



**v1.0.1**

- 增加数据初始化、数据查询页面

- 增加知识库页面，显示jupyter notebook


**v1.0.0**

可视化策略回测

- 支持在**本地运行可视化页面**, 仅需两行代码

- 已支持的**内置策略**: 目标权重策略, 风险平价策略 (策略不断扩充中)

- **我的策略**: 将回测策略及参数保存到个人空间

- **策略详情**: 查看策略详情, 并支持查看策略回测和模拟阶段绩效


## 客户端使用说明

```sh
创建虚拟环境(建议)
python -m venv venv_tgtrader

激活虚拟环境
1. windows
.\venv_tgtrader\Scripts\activate
2. mac/linux
source venv_tgtrader/bin/activate

安装tgtrader
pip install tgtrader

初始化数据
python -m tgtrader.streamlit_pages.init_data

启动客户端
1. 创建一个python文件(要与初始化数据时目录保持一致)，例如：tgtrader_cli.py
2. 在文件中添加以下代码：

from tgtrader.streamlit_pages.main import run
run()

3. 运行
  streamlit run tgtrader_cli.py

4. 定时任务启动(刷新财经新闻)
  python -m tgtrader.tasks.task_scheduler
```

#### 


    """)


from typing import Dict, Optional, Any
import signal
import sys
import logging
from datetime import datetime
from threading import Event
from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger
from apscheduler.jobstores.memory import MemoryJobStore
from tgtrader.dao.t_task import TTask
from tgtrader.dao.t_flow import FlowCfg
from tgtrader.service.flow_config_service import FlowConfigService
from loguru import logger

# 将APScheduler的日志转发到loguru
class InterceptHandler(logging.Handler):
    """将标准库的logging转发到loguru的处理器."""
    
    def emit(self, record: logging.LogRecord) -> None:
        # 获取对应的loguru级别
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # 找到调用者的文件名和行号
        frame, depth = logging.currentframe(), 2
        while frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(
            level, record.getMessage()
        )

# 配置APScheduler的日志
logging.getLogger("apscheduler").handlers = [InterceptHandler()]

class TaskScheduler:
    """
    任务调度器，负责管理和执行定时任务.
    
    使用APScheduler来调度任务，通过定期扫描任务表来管理任务的启动和停止.
    实现了单例模式，并处理系统信号以确保优雅关闭.
    
    Attributes:
        scheduler (BackgroundScheduler): APScheduler调度器实例
        running_jobs (Dict[int, str]): 正在运行的任务映射表，key为任务ID，value为job_id
    """
    
    _instance: Optional['TaskScheduler'] = None
    
    def __new__(cls) -> 'TaskScheduler':
        """单例模式实现."""
        if cls._instance is None:
            cls._instance = super(TaskScheduler, cls).__new__(cls)
        return cls._instance
    
    def __init__(self) -> None:
        """初始化任务调度器."""
        if not hasattr(self, 'initialized'):
            self.scheduler = BackgroundScheduler(
                jobstores={
                    'default': MemoryJobStore()
                },
                timezone='Asia/Shanghai'
            )
            self.running_jobs: Dict[int, str] = {}
            self._setup_signal_handlers()
            self.initialized = True
            
    def _setup_signal_handlers(self) -> None:
        """设置信号处理器."""
        signal.signal(signal.SIGINT, self._handle_shutdown)
        signal.signal(signal.SIGTERM, self._handle_shutdown)
        
    def _handle_shutdown(self, signum: int, frame: Optional[object]) -> None:
        """处理关闭信号.
        
        Args:
            signum (int): 信号编号
            frame (Optional[object]): 当前栈帧
        """
        logger.info(f"Received signal {signum}, shutting down...")
        self.stop()
        # 通知主线程退出
        if hasattr(self, '_exit_event'):
            self._exit_event.set()
        else:
            sys.exit(0)
        
    def start(self) -> None:
        """启动调度器并开始任务扫描."""
        # 启动调度器
        self.scheduler.start()
        logger.info("Task scheduler started")
        
        # 添加任务扫描job
        self.scheduler.add_job(
            func=self._scan_tasks,
            trigger='interval',
            seconds=5,
            id='task_scanner',
            name='Task Scanner',
            replace_existing=True
        )
        logger.info("Task scanner started")
        
    def stop(self) -> None:
        """停止调度器."""
        if self.scheduler.running:
            self.scheduler.shutdown()
            logger.info("Task scheduler stopped")
            
    def _scan_tasks(self) -> None:
        """扫描任务表，同步任务运行状态.
        
        该方法会定期检查数据库中的任务状态，并确保调度器中的任务与数据库保持同步.
        主要处理以下几种情况：
        1. 新增任务的调度
        2. 已删除任务的清理
        3. 任务配置（如crontab）的变更
        4. 任务运行状态的变更
        """
        try:
            # 获取所有任务
            all_tasks = TTask.select()
            
            # 获取当前运行的任务ID集合
            running_task_ids = set(self.running_jobs.keys())
            
            # 遍历所有任务
            for task in all_tasks:
                try:
                    self._sync_task_status(task, running_task_ids)
                except Exception as e:
                    logger.error(f"Failed to sync task {task.id}: {str(e)}")
                    continue
            
            # 清理已删除的任务
            self._cleanup_deleted_tasks(running_task_ids)
                    
        except Exception as e:
            logger.error(f"Failed to scan tasks: {str(e)}")
            
    def _sync_task_status(self, task: TTask, running_task_ids: set) -> None:
        """同步单个任务的状态.
        
        Args:
            task (TTask): 要同步的任务
            running_task_ids (set): 当前运行中的任务ID集合
        """
        if not self._validate_task(task):
            logger.warning(f"Task {task.id} validation failed, skipping")
            return
            
        if task.id not in running_task_ids:
            if task.status == 1:  # 运行状态
                # 数据库中是运行状态，但实际未运行的任务，启动它
                logger.info(f"Starting non-running task {task.id}")
                self._add_job(task)
        else:
            # 任务正在运行，首先检查任务是否应该继续运行
            if task.status == 0:  # 停止状态
                logger.info(f"Stopping running task {task.id} as it's marked as stopped")
                self._remove_job(task.id)
                return
                
            # 检查crontab是否变更
            job = self.scheduler.get_job(self.running_jobs[task.id])
            if not job:
                logger.warning(f"Job not found for task {task.id}, removing from running jobs")
                self._remove_job(task.id)
                return
                
            # 将当前任务的crontab转换为CronTrigger对象
            try:
                new_trigger = CronTrigger.from_crontab(task.crontab)
                current_trigger = job.trigger
                
                # 直接比较两个触发器的表达式
                if str(current_trigger) != str(new_trigger):
                    logger.info(f"Task {task.id} crontab changed from {current_trigger} to {task.crontab}")
                    self._remove_job(task.id)
                    self._add_job(task)
            except Exception as e:
                logger.exception(e)
                logger.error(f"Failed to compare crontab for task {task.id}: {str(e)}")
                return
                
    def _validate_task(self, task: TTask) -> bool:
        """验证任务配置的有效性.
        
        Args:
            task (TTask): 要验证的任务
            
        Returns:
            bool: 任务配置是否有效
        """
        try:
            if not task.crontab or not task.flow_id:
                logger.warning(f"Task {task.id} missing required fields")
                return False
                
            # 验证crontab格式
            CronTrigger.from_crontab(task.crontab)
            
            # 验证flow是否存在
            flow = FlowCfg.get_or_none(FlowCfg.flow_id == task.flow_id)
            if not flow:
                logger.warning(f"Flow {task.flow_id} not found for task {task.id}")
                return False
            return True
        except Exception as e:
            logger.exception(e)
            logger.warning(f"Task {task.id} validation failed: {str(e)}")
            return False
            
    def _cleanup_deleted_tasks(self, running_task_ids: set) -> None:
        """清理已删除的任务.
        
        Args:
            running_task_ids (set): 当前运行中的任务ID集合
        """
        for task_id in list(running_task_ids):
            try:
                TTask.get_by_id(task_id)
            except Exception as e:
                logger.exception(e)
                logger.info(f"Removing deleted task {task_id}: {str(e)}")
                self._remove_job(task_id)
            
    def _add_job(self, task: TTask) -> None:
        """添加任务到调度器.
        
        Args:
            task (TTask): 任务配置
        """
        if task.id in self.running_jobs:
            return
            
        try:
            # 创建cron触发器
            trigger = CronTrigger.from_crontab(task.crontab)
            
            # 添加任务到调度器
            job = self.scheduler.add_job(
                func=self._execute_task,
                trigger=trigger,
                args=[task.id],
                id=f"task_{task.id}",
                name=f"{task.flow_name}_{task.flow_id}",
                replace_existing=True
            )
            
            self.running_jobs[task.id] = job.id
            logger.info(f"Added task {task.flow_name} (ID: {task.id}) with schedule: {task.crontab}")
            
        except Exception as e:
            logger.exception(e)
            logger.error(f"Failed to add task {task.id}: {str(e)}")
            
    def _remove_job(self, task_id: int) -> None:
        """从调度器中移除任务.
        
        Args:
            task_id (int): 任务ID
        """
        job_id = self.running_jobs.get(task_id)
        if job_id:
            try:
                self.scheduler.remove_job(job_id)
                del self.running_jobs[task_id]
                logger.info(f"Removed task {task_id}")
            except Exception as e:
                logger.exception(e)
                logger.error(f"Failed to remove task {task_id}: {str(e)}")
                
    def _execute_task(self, task_id: int) -> None:
        """执行任务.
        
        Args:
            task_id (int): 任务ID
        """
        try:
            task = TTask.get_or_none(TTask.id == task_id)
            
            logger.info(f"Executing task {task.flow_name} (ID: {task.id}) at {datetime.now()}")
            
            FlowConfigService.run_flow(user=task.username, flow_id=task.flow_id)
            
        except Exception as e:
            logger.exception(e)
            logger.error(f"Failed to execute task {task_id}: {str(e)}")

    @classmethod
    def run_service(cls) -> None:
        """运行任务服务的主函数."""
        scheduler = cls()
        # 创建退出事件
        scheduler._exit_event = Event()
        scheduler.start()
        
        # 保持主线程运行
        try:
            # 等待退出事件
            scheduler._exit_event.wait()
        except (KeyboardInterrupt, SystemExit):
            scheduler.stop()

if __name__ == '__main__':
    TaskScheduler.run_service() 


# encoding: utf-8
from peewee import Model, SqliteDatabase
from tgtrader.utils.db_path_utils import get_common_data_database
import time

# 设置数据库路径
db = get_common_data_database()

class BaseModel(Model):
    class Meta:
        database = db

    @classmethod
    def create_or_update(cls, *args, **kwargs):
        instance = cls.get_or_none(**kwargs)
        if instance:
            instance.update_time = int(time.time() * 1000)
            instance.save()
        else:
            cls.create(**kwargs)



# encoding: utf-8
import time
from datetime import datetime
from tgtrader.dao.common import BaseModel, db
from peewee import AutoField, IntegerField, BigIntegerField, TextField, CompositeKey
from typing import List
from peewee import *
import re
from urllib.parse import urlparse
import requests
from loguru import logger


class TRssSource(BaseModel):
    # 主键自增id
    id = AutoField()
    # 用户名
    username = TextField()
    # rss源名称
    rss_name = TextField()
    # rss源url
    rss_url = TextField()
    # rss类型
    rss_type = TextField()
    # 创建时间
    create_time = BigIntegerField(
        default=lambda: int(time.time() * 1000))
    # 更新时间
    update_time = BigIntegerField(
        default=lambda: int(time.time() * 1000))

    MAX_NAME_LENGTH = 50
    MIN_NAME_LENGTH = 2

    class Meta:
        table_name = 't_rss_source'
        indexes = (
            (('username', 'rss_name'), True),
            (('username', 'rss_url'), True),
        )

    @classmethod
    def validate_rss_name(cls, rss_name: str) -> None:
        """验证RSS源名称"""
        if not rss_name:
            raise ValueError("RSS源名称不能为空")
        if len(rss_name) < cls.MIN_NAME_LENGTH:
            raise ValueError(f"RSS源名称长度不能小于{cls.MIN_NAME_LENGTH}个字符")
        if len(rss_name) > cls.MAX_NAME_LENGTH:
            raise ValueError(f"RSS源名称长度不能超过{cls.MAX_NAME_LENGTH}个字符")
        # 检查是否包含特殊字符
        if re.search(r'[<>"\']', rss_name):
            raise ValueError("RSS源名称不能包含特殊字符 < > ' \"")

    @classmethod
    def validate_rss_url(cls, rss_url: str) -> None:
        """验证RSS URL"""
        if not rss_url:
            raise ValueError("RSS URL不能为空")
        
        # 验证URL格式
        try:
            result = urlparse(rss_url)
            if not all([result.scheme, result.netloc]):
                raise ValueError("无效的URL格式")
            if result.scheme not in ['http', 'https']:
                raise ValueError("URL必须以http或https开头")
        except Exception as e:
            raise ValueError(f"无效的URL格式: {str(e)}")

    @classmethod
    def validate_rss_type(cls, rss_type: str) -> None:
        """验证RSS类型"""
        if not rss_type:
            raise ValueError("RSS类型不能为空")

    @classmethod
    def validate_all(cls, rss_name: str, rss_url: str, rss_type: str) -> None:
        """验证所有字段"""
        cls.validate_rss_name(rss_name)
        cls.validate_rss_url(rss_url)
        cls.validate_rss_type(rss_type)

    @classmethod
    def init_table(cls):
        with db:
            if not TRssSource.table_exists():
                db.create_tables([TRssSource])

    @classmethod
    def get_rss_sources(cls, username: str) -> List['TRssSource']:
        with db:
            return list(cls.select().where(cls.username == username))

    @classmethod
    def save_rss_source(cls, username: str, rss_name: str, rss_url: str, rss_type: str):
        try:
            # 验证所有字段
            cls.validate_all(rss_name, rss_url, rss_type)
            
            with db:
                cls.create(
                    username=username,
                    rss_name=rss_name,
                    rss_url=rss_url,
                    rss_type=rss_type
                )
        except ValueError as e:
            raise Exception(str(e))
        except IntegrityError:
            raise Exception(f"RSS源已存在，请修改RSS源名称或URL")
        except Exception as e:
            raise Exception(f"保存RSS源失败: {str(e)}")
    
    @classmethod
    def delete_rss_source(cls, _id: int):
        with db:
            cls.delete().where(cls.id == _id).execute()

    @classmethod
    def update_rss_source(cls, _id: int, rss_name: str, rss_url: str, rss_type: str):
        try:
            # 验证所有字段
            cls.validate_all(rss_name, rss_url, rss_type)
            
            with db:
                cls.update(
                    rss_name=rss_name,
                    rss_url=rss_url,
                    rss_type=rss_type,
                    update_time=int(time.time() * 1000)
                ).where(cls.id == _id).execute()
        except ValueError as e:
            raise Exception(str(e))
        except IntegrityError:
            raise Exception(f"RSS源已存在，请修改RSS源名称或URL")
        except Exception as e:
            raise Exception(f"更新RSS源失败: {str(e)}")

TRssSource.init_table()



# encoding: utf-8

import duckdb
import pandas as pd
from typing import List, Optional, Union, Dict

class DuckDBQueryDF:
    """
    A utility class for querying pandas DataFrames using DuckDB.
    
    This class provides methods to execute SQL queries against pandas DataFrames
    using DuckDB as the query engine. It supports both single DataFrame queries
    and queries involving multiple DataFrames.
    """
    
    def __init__(self, dfs: Optional[Dict[str, pd.DataFrame]] = None) -> None:
        """
        Initialize a new DuckDB connection for DataFrame querying.
        
        Args:
            dfs: Optional dictionary of pandas DataFrames to query. The keys will be used as table names.
        """
        self.con = duckdb.connect(database=':memory:')
        if dfs is not None:
            self.register_multiple(dfs)
    
    def register_multiple(self, dfs: Dict[str, pd.DataFrame]) -> None:
        """
        Register multiple DataFrames as tables for querying.
        
        Args:
            dfs: A dictionary where keys are table names and values are pandas DataFrames.
        """
        for name, df in dfs.items():
            self.con.register(name, df)
    
    def query(self, sql: str) -> pd.DataFrame:
        """
        Execute a SQL query against the registered DataFrame.
        
        Args:
            sql: SQL query string to execute
            
        Returns:
            pd.DataFrame: Result of the SQL query as a pandas DataFrame
            
        Examples:
            >>> df = pd.DataFrame({'a': [1, 2, 3], 'b': ['x', 'y', 'z']})
            >>> dq = DuckDBQueryDF(df)
            >>> result = dq.query("SELECT * FROM df WHERE a > 1")
        """
        return self.con.execute(sql).df()
    
    def close(self) -> None:
        """
        Close the DuckDB connection.
        """
        self.con.close()
    
    def __enter__(self) -> 'DuckDBQueryDF':
        """
        Context manager entry point.
        
        Returns:
            DuckDBQueryDF: Self reference for use in context manager
        """
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Context manager exit point.
        
        Args:
            exc_type: Exception type if an exception occurred
            exc_val: Exception value if an exception occurred
            exc_tb: Exception traceback if an exception occurred
        """
        self.close()
